import inspect
import os
import re

from pytest_check import check

from . import services
from .result_formatter import ResultFormatter
from .result_summary import ResultSummary
from .utils import Utils
from medver_pytest.lib.view import Gui


# -------------------
## Holds the verifyxx() functions
class Verifier:
    # -------------------
    ## constructor
    def __init__(self):
        pass

    # -------------------
    ## returns a copy of the default formatter
    #
    # @return None
    def get_formatter(self):
        # TODO return a copy of the default formatter
        return ResultFormatter()

    # -------------------
    ## report a verification abort.
    # Note: currently on in manual tests.
    #
    # @param msg       message to add to report
    # @param location  the location of the abort
    # @return None
    def report_abort(self, msg, location):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(None, None)
        if location is None:
            location = Utils.get_location(levels=4)
        rs.location = location

        self._set_actual(f'aborted: {msg}', rs)
        self._set_expected('successful completion', rs)

        self._handle_fail(rs)

    # -------------------
    ## verify the actual value is True
    #
    # @param actual       the actual value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify(self, actual, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_expected(True, rs)

        if actual:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is True
    # alias for self.verify()
    #
    # @param actual       the actual value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_true(self, actual, reqids=None, formatter=None):
        # note: even though the code is identical,
        # do not call self.verify(), otherwise the location is incorrect
        # the call to _create_result_summary() must be local to this function

        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_expected(True, rs)

        if actual:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is False
    #
    # @param actual       the actual value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_false(self, actual, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_expected(False, rs, is_not=False)

        if not actual:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value equals the expected value
    #
    # @param expected     the expected value
    # @param actual       the actual value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_equal(self, expected, actual, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_expected(expected, rs)

        if actual == expected:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value does not equal the expected value
    #
    # @param expected     the expected value
    # @param actual       the actual value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_not_equal(self, expected, actual, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_expected(expected, rs, is_not=True)

        if actual != expected:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is None
    #
    # @param actual       the actual value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_none(self, actual, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_expected(None, rs)

        if actual is None:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is None
    #
    # @param actual       the actual value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_is_none(self, actual, reqids=None, formatter=None):
        # note: even though the code is identical,
        # do not call self.verify_none(), otherwise the location is incorrect
        # the call to _create_result_summary() must be local to this function

        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_expected(None, rs)

        if actual is None:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is not None
    #
    # @param actual       the actual value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_not_none(self, actual, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_expected(None, rs, is_not=True)

        if actual is not None:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is in the given list
    #
    # @param actual       the actual value
    # @param exp_list     the list of values
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_in(self, actual, exp_list, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_expected(exp_list, rs, prefix='in ')

        if actual in exp_list:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is not in the given list
    #
    # @param actual       the actual value
    # @param exp_list     the list of values
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_not_in(self, actual, exp_list, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_expected(exp_list, rs, prefix='in ', is_not=True)

        if actual not in exp_list:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the left value is less than the right value
    #
    # @param left         the left value
    # @param right        the right value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_lt(self, left, right, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(left, rs)
        self._set_expected(right, rs, prefix='less than ')

        if left < right:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the left value is less than or equal to the right value
    #
    # @param left         the left value
    # @param right        the right value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_le(self, left, right, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(left, rs)
        self._set_expected(right, rs, prefix='less than or equal to ')

        if left <= right:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the left value is greater than the right value
    #
    # @param left         the left value
    # @param right        the right value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_gt(self, left, right, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(left, rs)
        self._set_expected(right, rs, prefix='greater than ')

        if left > right:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the left value is greater than or equal to the right value
    #
    # @param left         the left value
    # @param right        the right value
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_ge(self, left, right, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(left, rs)
        self._set_expected(right, rs, prefix='greater or equal to ')

        if left >= right:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value matches the expected regex pattern
    #
    # @param actual       the actual value
    # @param regex        the regex pattern
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_regex(self, actual, regex, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_regex_expected(regex, rs)

        if re.search(regex, actual):
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value does not matche the expected regex pattern
    #
    # @param actual       the actual value
    # @param regex        the regex pattern
    # @param reqids       (optional) one or more reqids associated with this test
    # @param formatter    (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_not_regex(self, actual, regex, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_regex_expected(regex, rs, is_not=True)

        if not re.search(regex, actual):
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is within the expected value +/- the given tolerance
    #
    # @param expected       the expected value
    # @param actual         the actual value
    # @param abs_tolerance  numeric; the absolute +/- tolerance
    # @param reqids         (optional) one or more reqids associated with this test
    # @param formatter      (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_delta(self, expected, actual, abs_tolerance, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        # ok if abs_tolerance == 0, behaves like verify_equal()
        # fail if the tolerance is negative
        if abs_tolerance < 0:
            services.abort(f'invalid abs_tolerance, must be >= 0; at {rs.location}')
            return

        self._set_actual(actual, rs)
        self._set_delta_expected(expected, rs, abs_tolerance)

        if abs(actual - expected) <= abs_tolerance:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is outside the expected value +/- the given tolerance
    #
    # @param expected       the expected value
    # @param actual         the actual value
    # @param abs_tolerance  numeric; the absolute +/- tolerance
    # @param reqids         (optional) one or more reqids associated with this test
    # @param formatter      (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_not_delta(self, expected, actual, abs_tolerance, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        # ok if abs_tolerance == 0, behaves like verify_equal()
        # fail if the tolerance is negative
        if abs_tolerance < 0:
            services.abort(f'invalid abs_tolerance, must be >= 0; at {rs.location}')
            return

        self._set_actual(actual, rs)
        self._set_delta_expected(expected, rs, abs_tolerance, is_not=True)

        if abs(actual - expected) > abs_tolerance:
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is within the expected value +/- the given percentage tolerance
    #
    # @param expected       the expected value
    # @param actual         the actual value
    # @param rel_tolerance  numeric; the relative percentage +/- % tolerance
    # @param reqids         (optional) one or more reqids associated with this test
    # @param formatter      (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_delta_pct(self, expected, actual, rel_tolerance, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        # ok if rel_tolerance == 0, behaves like verify_equal()
        # fail if the percentage is negative
        if rel_tolerance < 0:
            services.abort(f'invalid rel_tolerance, must be >= 0; at {rs.location}')
            return

        self._set_actual(actual, rs)
        self._set_delta_expected(expected, rs, rel_tolerance, is_pct=True)

        if abs(actual - expected) <= abs((rel_tolerance / 100.0) * expected):
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## verify the actual value is outside the expected value +/- the given percentage tolerance
    #
    # @param expected       the expected value
    # @param actual         the actual value
    # @param rel_tolerance  numeric; the relative percentage +/- % tolerance
    # @param reqids         (optional) one or more reqids associated with this test
    # @param formatter      (optional) formatter object to assist in formatting the incoming values
    # @return None
    def verify_not_delta_pct(self, expected, actual, rel_tolerance, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        self._set_actual(actual, rs)
        self._set_delta_expected(expected, rs, rel_tolerance, is_pct=True, is_not=True)

        # ok if rel_tolerance == 0, behaves like verify_equal()
        # fail if the percentage is negative
        if rel_tolerance < 0:
            services.abort(f'invalid rel_tolerance, must be >= 0; at {rs.location}')
            return

        if abs(actual - expected) > abs((rel_tolerance / 100.0) * expected):
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

    # -------------------
    ## manual verify the actual value matches the expected value
    #
    # @param verify_desc    the tester instructions to do the verication
    # @param expected       the expected value
    # @param reqids         (optional) one or more reqids associated with this test
    # @param formatter      (optional) formatter object to assist in formatting the incoming values
    # @return None
    def manual_verify(self, verify_desc, expected, reqids=None, formatter=None):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        rs = self._create_result_summary(reqids, formatter)

        if not services.view:
            services.view = Gui()

        state, actual = services.view.show_verify(services.proto.protocol_id, verify_desc, expected)
        if state == 'abort':
            msg = 'manual_verify: tester pressed Abort'
            services.harness.ver.report_abort(msg, rs.location)
            services.abort(msg)

        services.proto.update_tester_initials()

        self._set_actual(actual, rs)
        self._set_expected(expected, rs)

        if state == 'pass':
            self._handle_pass(rs)
        else:
            self._handle_fail(rs)

        return actual

    # -------------------
    ## the verify passed, so save the result
    #
    # @param rs    result summary object
    # @return None
    def _handle_pass(self, rs):
        self._save_result_pass(rs)

    # -------------------
    ## the verify failed, so save the result
    # Note: in IUV mode does nothing.
    # In normal mode gathers the failure information and continues
    #
    # @param rs    result summary object
    # @return None
    def _handle_fail(self, rs):
        __tracebackhide__ = True  # pylint: disable=unused-variable
        self._save_result_fail(rs)

        msg = self._format_fail_msg(rs)

        if services.cfg.iuvmode:  # pragma: no cover
            # coverage: iuvmode always set during IUV and UT runs
            pass
        else:
            # TODO stop using pytest-check
            with check:  # pragma: no cover
                # coverage: iuvmode set during IUV and UT runs; never gets here
                check.set_no_tb()
                raise AssertionError(f'at {rs.location}{msg}\n{self._get_trace()}')

    # -------------------
    ## format the traceback when an error occurs
    #
    # @return a traceback list
    def _get_trace(self):  # pragma: no cover
        # coverage: traceback not used in IUV

        # no point going farther back than 3 levels
        #    get_trace
        #    get_full_context
        level = 2
        lines = ['   --']
        while True:
            file, line, func, context = self._get_traceback(level)
            if file is None:
                break

            # uncomment to debug
            # lines.append(f'   [{level}] found {file} {line} {func} {context}')

            level += 1
            # this starts it in unitest, might as well stop
            if '_callTestMethod' in func:
                break
            if 'medver_pytest' in file:
                continue
            if 'miniconda3' in file:
                continue
            if 'site-packages' in file:
                continue
            if '<module>' in func:
                continue
            line = f'   {file}:{line} in {func}() -> {context}'
            lines.append(line)
        return '\n'.join(reversed(lines))

    # -------------------
    ## get traceback information at the given level
    #
    # @param level   the traceback level to report on
    # @return a traceback tuple: filename, line number, function name, context
    def _get_traceback(self, level):  # pragma: no cover
        # coverage: traceback not used in IUV
        try:
            (_, filename, line, funcname, contextlist) = inspect.stack()[level][0:5]
        except IndexError:
            return None, None, None, None

        try:
            filename = os.path.relpath(filename)
        except ValueError:
            filename = os.path.abspath(filename)
        context = contextlist[0].strip() if contextlist else ""
        return filename, line, funcname, context

    # -------------------
    ## save a pass result using the given result summary
    #
    # @param rs    result summary object
    # @return None
    def _save_result_pass(self, rs):
        rs.passed()
        services.proto.add_result(rs)

    # -------------------
    ## save a fail result using the given result summary
    #
    # @param rs    result summary object
    # @return None
    def _save_result_fail(self, rs):
        rs.failed()
        services.proto.add_result(rs)

    # -------------------
    ## create a result summary
    #
    # @param reqids      the requirements impacted by this result
    # @param formatter   the formatter, if any, to use for formatting values
    # @return the result summary object
    def _create_result_summary(self, reqids, formatter):
        location = Utils.get_location()

        # create a results summary and pre-populate it
        rs = ResultSummary()
        rs.location = location
        rs.set_reqids(reqids)
        rs.formatter = formatter
        return rs

    # -------------------
    ## save the actual value to the result summary
    #
    # @param actual   the actual value
    # @param rs       result summary object
    # @return None
    def _set_actual(self, actual, rs):
        rs.format_actual(actual)

    # -------------------
    ## save a formatted version of the expected value to the result summary
    #
    # @param expected   the expected value
    # @param rs         result summary object
    # @param prefix     (optional) a prefix string to use in the formatted value
    # @param suffix     (optional) a suffix string to use in the formatted value
    # @param is_not     (optional) if True, use "not" in the formatter value
    # @return None
    def _set_expected(self, expected, rs, prefix=None, suffix=None, is_not=False):
        rs.format_expected(expected, prefix, suffix, is_not=is_not)

    # -------------------
    ## save a formatted version of the expected regex value to the result summary
    #
    # @param expected   the expected value
    # @param rs         result summary object
    # @param is_not     (optional) if True, use "not" in the formatter value
    # @return None
    def _set_regex_expected(self, expected, rs, is_not=False):
        rs.format_regex_expected(expected, is_not=is_not)

    # -------------------
    ## save a formatted version of the expected delta value to the result summary
    #
    # @param expected   the expected value
    # @param rs         result summary object
    # @param delta      the +/- delta value (may be a percentage)
    # @param is_pct     (optional) delta value is percentage
    # @param is_not     (optional) if True, use "not" in the formatter value
    # @return None
    def _set_delta_expected(self, expected, rs, delta, is_pct=False, is_not=False):
        rs.format_delta_expected(expected, delta, is_pct=is_pct, is_not=is_not)

    # -------------------
    ## generate a formatted fail message based on the result summary content
    #
    # @param rs         result summary object
    # @return None
    def _format_fail_msg(self, rs):
        t = f'({rs.expected_type})'
        line1 = f'Expected {t: <10}: {rs.expected_formatted}'
        t = f'({rs.actual_type})'
        line2 = f'Actual   {t: <10}: {rs.actual_formatted}'
        if services.cfg.iuvmode:  # pragma: no cover
            # coverage: iuvmode is only set during IUV and UT runs
            services.harness.iuv.print(line1)
            services.harness.iuv.print(line2)

        msg = f'\n   {line1}\n   {line2}'
        return msg



from medver_pytest import services
from medver_pytest.constants import Constants
from medver_pytest.utils import Utils


# -------------------
## Base class for generating an HTML with no meta, style, etc, tags
# useful for embedding the html into other HTML files
class GenBaseMinHtml:  # pylint: disable=too-few-public-methods

    # -------------------
    ## constructor
    def __init__(self):
        ## holds file pointer for the generated text file
        self._fp = None

    # -------------------
    ## generate test run information
    #
    # @return None
    def _gen_test_run_details(self):
        self._gen_title('Test Run Details')
        self._fp.write('<ul>\n')
        self._gen_li(f'Test Run Type: {services.cfg.test_run_type}')
        self._gen_li(f'Test Run ID: {services.cfg.test_run_id}')
        dts = Utils.get_dts(use_cfg_fmt=True)
        self._gen_li(f'Document Generated: {dts}')
        self._gen_li(f'medver-pytest version: v{Constants.version}')
        self._fp.write('</ul>\n')

    # -------------------
    ## generate title
    #
    # @param title  the title to draw
    # @param arg_id   option id= tag for the h3
    # @return None
    def _gen_title(self, title, arg_id=None):
        self._gen_h3(title, arg_id)

    # -------------------
    ## generate an h3 line
    #
    # @param text     the text to draw
    # @param arg_id   option id= tag for the h3
    # @return None
    def _gen_h3(self, text, arg_id=None):
        if arg_id is None:
            self._fp.write(f'<h3>{text}</h3>\n')
        else:
            self._fp.write(f'<h3 id="{arg_id}">{text}</h3>\n')

    # -------------------
    ## generate an paragraph highlighted and optionally indented
    #
    # @param text    the text to draw
    # @param indent  if true the text is indented
    # @return None
    def _gen_p_strong(self, text, indent=False):
        indent_str = ''
        if indent:
            indent_str = ' style="padding-left: 30px"'
        self._fp.write(f'<p{indent_str}><strong>{text}</strong></p>\n')

    # -------------------
    ## generate an li line
    #
    # @param text  the text to draw
    # @return None
    def _gen_li(self, text):
        self._fp.write(f'    <li>{text}</li>\n')

    # -------------------
    ## generate a table open tag
    #
    # @param css_class   optional CSS class to use for this table
    # @return None
    def _gen_table_open(self, css_class=None):
        if css_class is None:
            self._fp.write('<table>\n')
        else:
            self._fp.write(f'<table class="{css_class}">\n')

    # -------------------
    ## generates a end of table tag
    #
    # @returns None
    def _gen_table_close(self):
        self._fp.write('</table>\n')

    # -------------------
    ## generates a start of table heading section
    #
    # @returns None
    def _gen_thead_open(self):
        self._fp.write('    <thead>\n')

    # -------------------
    ## generates a end of table head tag
    #
    # @returns None
    def _gen_thead_close(self):
        self._fp.write('    </thead>\n')

    # -------------------
    ## generates a table body open definition
    #
    # @return None
    def _gen_tbody_open(self):
        self._fp.write('    <tbody>\n')

    # -------------------
    ## generates a end of table body tag
    #
    # @returns None
    def _gen_tbody_close(self):
        self._fp.write('    </tbody>\n')

    # -------------------
    ## generates a table row definition with the given class
    #
    # @param css_class  the CSS class to use
    # @returns None
    def _gen_tr_open(self, css_class=None):
        if css_class is None:
            self._fp.write('    <tr>\n')
        else:
            self._fp.write(f'    <tr class="{css_class}">\n')

    # -------------------
    ## generates a end of row tag
    #
    # @returns None
    def _gen_tr_close(self):
        self._fp.write('    </tr>\n')

    # -------------------
    ## generates a heading column definition
    #
    # @param text       the text to put in the title
    # @returns None
    def _gen_th_strong(self, text=None):
        if text is None:
            self._fp.write('        <th></th>\n')
        else:
            self._fp.write(f'        <th><strong>{text}</strong></th>\n')

    # -------------------
    ## generates a table heading definition with the given class
    #
    # @param text       the text to put in the title
    # @param css_class  the CSS class to use
    # @returns None
    def _gen_th(self, text=None, css_class=None):
        css_str = ''
        if css_class is not None:
            css_str = f' class="{css_class}"'

        if text is None:
            self._fp.write(f'        <th{css_str}></th>\n')
        else:
            self._fp.write(f'        <th{css_str}>{text}</th>\n')

    # -------------------
    ## generates a table column definition with the given class and highlighted
    #
    # @param text       the text to put in the title
    # @param css_class  the CSS class to use
    # @returns None
    def _gen_td_strong(self, text, css_class):
        if text is None:
            self._fp.write(f'        <td class="{css_class}"></td>\n')
        else:
            self._fp.write(f'        <td class="{css_class}"><strong>{text}</strong></td>\n')

    # -------------------
    ## generates a table column definition with the given class and optionally indented
    #
    # @param text       the text to put in the title
    # @param css_class  the CSS class to use
    # @param indent     if true, the text is indented
    # @returns None
    def _gen_td(self, text, css_class, indent=False):
        indent_str = ''
        if indent:
            indent_str = ' style="padding-left: 30px"'
        if text is None:
            self._fp.write(f'        <td class="{css_class}"{indent_str}></td>\n')
        else:
            self._fp.write(f'        <td class="{css_class}"{indent_str}>{text}</td>\n')

    # -------------------
    ## generates a table column definition indented and spanning multiple columns
    #
    # @param text   the text to put in the title
    # @param cols   the number of columns to span
    # @returns None
    def _gen_td_indented_spanned(self, text, cols):
        self._fp.write(f'<td colspan="{cols}" style="padding-left: 30px">{text}</td>\n')

    # -------------------
    ## generates a line break
    #
    # @returns None
    def _gen_blank_line(self):
        self._fp.write('<br/>\n')



import tkinter as tk
import tkinter.scrolledtext as tkst

from .dlg_base import DlgBase
from medver_pytest import services


# -------------------
## holds Dialog Box for a Manual Verify
class DlgManualVerify(DlgBase):
    # -------------------
    ## constructor
    def __init__(self):
        ## see base class for doc
        self._dlg = None

        super().__init__()

        ## the return state, holds abort, pass, or fail
        self._return_value = 'initial'
        ## the actual value that occurred, cannot be ''
        self._actual = ''
        ## whether actual was ever entered or not
        self._actual_entered = False

        # widgets
        ## the Pass btn
        self._btn_pass = None
        ## the Fail btn
        self._btn_fail = None
        ## the Actual Text widget
        self._txt_actual = None

    # -------------------
    ## show dialog for a manual verify
    #
    # @param parent   the tk root window
    # @param title    the dlg title
    # @param verify_desc   the action for the tester to perform
    # @param expected      the expected verification value
    # @return the final state of the dlg (done, fail, abort), the actual verification value
    def manual_verify(self, parent, title, verify_desc, expected):
        ## see base class for doc
        self._parent = parent
        self._show_dlg_manual_verify(title, verify_desc, expected)
        self._parent.wait_window(self._dlg)
        return self._return_value, self._actual

    # -------------------
    ## show the dlg for performing a manual action
    #
    # @param title   the dlg title
    # @param verify_desc   the action for the tester to perform
    # @param expected      the expected verification value
    # @return None
    def _show_dlg_manual_verify(self, title, verify_desc, expected):
        row = 0
        msg = 'Perform verification, enter Actual, then press Pass or Fail'
        row = self._dlg_common_first(row, title, msg, self._ensure_initials)

        row += 1
        self._create_verify_desc(row, verify_desc)
        services.logger.user(f'verify: {verify_desc}')

        row += 1
        self._create_cell_lbl('Expected:', row, 0)
        row += 1
        self._create_expected(row, expected)
        services.logger.user(f'expected: {expected}')

        row += 1
        self._create_cell_lbl('Actual:', row, 0)
        row += 1
        self._create_actual(row)

        row += 1
        self._create_empty_row(row)

        row += 1
        self._create_pass_btn(row, 0)
        self._create_fail_btn(row, 1)

        ## set callback for handling a click and to ensure intials are set
        self._dlg_common_last(row,
                              self._handle_click,
                              self._ensure_initials)

    # -------------------
    ## the verification action for the tester to perform
    #
    # @param row          the row to place it in
    # @param verify_desc  the verification the tester needs to perform
    # @return None
    def _create_verify_desc(self, row, verify_desc):
        lbl = tkst.ScrolledText(self._frame,
                                wrap=tk.WORD,  # wrap text on a word boundary
                                height=5,  # lines
                                width=self.col0_txt_width,
                                font=self.common_font,
                                fg='black',
                                bg='lightgrey',
                                highlightbackground='black', highlightthickness=2)
        lbl.insert(1.0, verify_desc)
        lbl.configure(state='disabled')  # make it readonly
        lbl.grid(row=row, column=0, sticky='NSWE', columnspan=3)

    # -------------------
    ## the expected value from the verifcation action
    #
    # @param row       the row to place it in
    # @param expected  the expected value of the verification
    # @return None
    def _create_expected(self, row, expected):
        lbl = tkst.ScrolledText(self._frame,
                                wrap=tk.WORD,  # wrap text on a word boundary
                                height=5,  # lines
                                width=self.col0_txt_width,
                                font=self.common_font,
                                fg='black',
                                bg='lightgrey',
                                highlightbackground='black', highlightthickness=2)
        lbl.insert(1.0, expected)
        lbl.configure(state='disabled')  # make it readonly
        lbl.grid(row=row, column=0, sticky='NSWE', columnspan=3)

    # -------------------
    ## btn indicates the tester passed the verification
    #
    # @param row     the row to place it in
    # @param col     the column to place it in
    # @return None
    def _create_pass_btn(self, row, col):
        self._btn_pass = tk.Button(self._frame, text='Pass',
                                   command=lambda: self._handle_click('pass'),
                                   width=self.col0_btn_width,
                                   font=self.common_font,
                                   fg='black',
                                   bg='lightgreen',
                                   disabledforeground='lightgrey',
                                   highlightbackground='lightgrey', highlightthickness=2)
        self._btn_pass.grid(row=row, column=col, sticky='W')

    # -------------------
    ## disable the Done button
    #
    # @return None
    def _pass_disable(self):
        self._btn_pass['state'] = 'disabled'

    # -------------------
    ## enable the Done button
    #
    # @return None
    def _pass_enable(self):
        self._btn_pass['state'] = 'normal'

    # -------------------
    ## place a highlight around the Done btn
    #
    # @return None
    def _pass_highlight(self):
        self._btn_pass['fg'] = 'black'
        self._btn_pass['highlightbackground'] = 'black'
        self._btn_pass['highlightthickness'] = 2

    # -------------------
    ## unhighlight the Done btn
    #
    # @return None
    def _pass_unhighlight(self):
        self._btn_pass['fg'] = 'black'
        self._btn_pass['highlightbackground'] = 'lightgrey'
        self._btn_pass['highlightthickness'] = 2

    # -------------------
    ## btn that indicates the tester failed the verififcation
    #
    # @param row     the row to place it in
    # @param col     the column to place it in
    # @return None
    def _create_fail_btn(self, row, col):
        self._btn_fail = tk.Button(self._frame, text='Fail',
                                   command=lambda: self._handle_click('fail'),
                                   width=self.col0_btn_width,
                                   font=self.common_font,
                                   fg='black',
                                   bg='gold',
                                   disabledforeground='lightgrey',
                                   highlightbackground='lightgrey', highlightthickness=2)
        self._btn_fail.grid(row=row, column=col, sticky='W')

    # -------------------
    ## disable the Can't Do button
    #
    # @return None
    def _fail_disable(self):
        self._btn_fail['state'] = 'disabled'

    # -------------------
    ## enable the Can't Do button
    #
    # @return None
    def _fail_enable(self):
        self._btn_fail['state'] = 'normal'

    # -------------------
    ## highlight the Can't Do btn
    #
    # @return None
    def _fail_highlight(self):
        self._btn_fail['fg'] = 'black'
        self._btn_fail['highlightbackground'] = 'black'
        self._btn_fail['highlightthickness'] = 2

    # -------------------
    ## unhighlight the Can't Do button
    #
    # @return None
    def _fail_unhighlight(self):
        self._btn_fail['fg'] = 'black'
        self._btn_fail['highlightbackground'] = 'lightgrey'
        self._btn_fail['highlightthickness'] = 2

    # -------------------
    ## Create the Actual text area. The tester fills
    # it in to indicate the actual value
    #
    # @param row     the row to place it in
    # @return None
    def _create_actual(self, row):
        height = 5
        self._txt_actual = tkst.ScrolledText(self._frame,
                                             wrap=tk.WORD,  # wrap text on a word boundary
                                             height=height,  # lines
                                             width=self.col0_txt_width,
                                             font=self.common_font,
                                             fg='black',
                                             bg='white',
                                             highlightbackground='lightgrey', highlightthickness=2)
        self._txt_actual.configure(state='disabled')  # initially, make it readonly
        self._txt_actual.focus()
        ## set callback for key release
        self._txt_actual.bind('<KeyRelease>', self._actual_key_pressed)
        self._txt_actual.grid(row=row, column=0, sticky='NSWE', columnspan=3)

    # -------------------
    ## handle an entry in the Actual box
    #
    # @param unused   ignored; the key that was pressed
    # @return None
    # noinspection PyUnusedLocal
    def _actual_key_pressed(self, unused):  # pylint: disable=unused-argument
        self._actual_entered = True
        self._get_actual()
        self._check_next()

    # -------------------
    ## get the content form the actual text box
    #
    # @return None
    def _get_actual(self):
        # get the actual text
        self._actual = self._txt_actual.get('1.0', 'end')
        self._actual = self._actual.strip()

    # -------------------
    ## check content of actual field;
    # should be 1 or more characters
    #
    # @return True if _actual has invalid content, False otherwise
    def _actual_is_wrong(self):
        return len(self._actual) < 1

    # -------------------
    ## disable the actual text box
    #
    # @return None
    def _actual_disable(self):
        self._txt_actual['state'] = 'disabled'
        self._txt_actual['fg'] = 'black'

    # -------------------
    ## enable the actual text box
    # note: don't clear out actual text in case the tester wants to extend it
    #
    # @return None
    def _actual_enable(self):
        self._txt_actual['state'] = 'normal'
        self._txt_actual['fg'] = 'black'

    # -------------------
    ## handle a dlg box click
    #
    # @param option  the btn the tester clicked
    # @return None
    def _handle_click(self, option):
        self._get_actual()
        if option == 'pass':
            services.logger.user('click Pass')
            self._actual_enable()
            self._fail_unhighlight()
            self._pass_highlight()
            self._return_value = 'pass'
            self._check_next()

        elif option == 'fail':
            services.logger.user('click Fail')
            self._actual_enable()
            self._fail_highlight()
            self._pass_unhighlight()
            self._return_value = 'fail'
            self._check_next()

        elif option == 'next':
            services.logger.user(f'actual: {self._actual}')
            services.logger.user('click Next')

            ok = self._check_next()
            if ok:
                self._dlg.destroy()

        elif option == 'abort':
            services.logger.user(f'actual: {self._actual}')
            services.logger.user('click Abort')
            self._return_value = 'abort'
            self._dlg.destroy()

    # -------------------
    ## check if Next button should be enabled or not.
    # Set message box as appropriate.
    #
    # @return True if all is ready, False otherwise
    def _check_next(self):
        ok = False
        if self._return_value == 'initial' and not self._actual_entered:
            self._next_disable()
            self._set_message('red', 'Perform verification, enter Actual, then press Pass or Fail')
        elif self._return_value == 'initial' and self._actual_entered:
            # assume that if any value was ever entered in Actual, that the verification was performed
            self._set_message('red', 'Enter Actual, then press Pass or Fail')
        elif self._return_value == 'initial' and not self._actual_is_wrong():
            # actual was entered, but pass or fail wasn't pressed
            self._next_disable()
            self._set_message('red', 'Check Actual, then press Pass or Fail')
        elif self._return_value != 'initial' and self._actual_is_wrong():
            # either Pass or Fail was pressed
            self._next_disable()
            self._set_message('red', 'Actual is too short. Enter Actual value.')
        else:
            # actual is filled in and pass/fail was clicked
            self._next_enable()
            self._set_message('green', f'{self._return_value.capitalize()} selected. '
                                       'Check Actual, press Next when ready')
            ok = True
        return ok

    # -------------------
    ## disable buttons (except abort) if initials box is empty
    #
    # @return None
    def _ensure_initials(self):
        if self._tester_initials_is_wrong():
            self._actual_disable()
            self._pass_disable()
            self._fail_disable()
            self._next_disable()
            self._set_message('red', 'Enter your initials (2 or 3 characters)')
        else:
            self._actual_enable()
            self._pass_enable()
            self._fail_enable()
            self._check_next()



from medver_pytest import services


# -------------------
## holds Header information
class Header:
    # -------------------
    ## constructor
    def __init__(self):
        ## holds content for left side of header
        self.left = ''

        ## holds content for center section of header
        self.middle = ''

        ## holds content for right side of header
        self.right = ''

    # -------------------
    ## save a header item
    #
    # @param item    the name of the item to set: left, middle or right
    # @param value   the value to set
    # @return None
    def __setitem__(self, item, value):
        if item == 'left':
            self.left = value
            return

        if item == 'middle':
            self.middle = value
            return

        if item == 'right':
            self.right = value
            return

        services.abort(f'bad item name: {item}')

    # -------------------
    ## get a value from this object
    #
    # @param item   the name of the attribute to get
    # @return the value of the attribute
    def __getitem__(self, item):
        if item == 'left':
            return self.left

        if item == 'middle':
            return self.middle

        if item == 'right':
            return self.right

        services.abort(f'bad item name: {item}')
        return None



from .gen_tp_base import GenTpBase
from ..common.gen_base_minhtml import GenBaseMinHtml
from medver_pytest import services
from medver_pytest.result_summary import ResultSummary


# -------------------
## Generates a Test Protocol report in min-html format
class GenTpMinHtml(GenBaseMinHtml, GenTpBase):
    # -------------------
    ## constructor
    #
    # @param protocols   the data to use
    # @param do_results  generate results or not (default True)
    def __init__(self, protocols, do_results=True):
        ## see base clase for doc
        self._doc_path = None

        GenBaseMinHtml.__init__(self)
        GenTpBase.__init__(self, protocols, do_results)

        self._init('min-html', protocols, do_results)

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()
        # uncomment to debug:
        # from medver_pytest.lib.stdout_logger import StdoutLogger as logger
        # logger.dbg(f'protocols: {json.dumps(self._protocols, indent=2)}')

        # TODO add data from self._requirements = Requirements.fast_load()

        with open(self._doc_path, 'w', encoding='UTF-8') as self._fp:
            self._gen_test_run_details()
            title = self._get_title()
            self._gen_title(title)

            for _, protocol in self._protocols.items():
                self._gen_protocol(protocol)

    # -------------------
    ## generate a protocol table using the given data
    #
    # @param protocol   the data to use
    # @return None
    def _gen_protocol(self, protocol):
        ## see GenTpBase
        self._testers = []
        self._gen_protocol_info(protocol)

        requirements = []
        self._gen_steps_info(protocol['steps'], requirements)
        self._gen_reqmt_table(requirements)
        if self._testers:
            self._gen_blank_line()
            self._gen_testers()
        self._gen_blank_line()
        self._gen_blank_line()

    # -------------------
    ## generates the start of the steps table
    #
    # @returns None
    def _gen_steps_table_open(self):
        self._gen_table_open('ver-report-table')
        self._gen_thead_open()
        self._gen_tr_open(css_class='ver-report-steps-hdg')
        self._gen_th('Step', 'ver-report-steps-hdg-col1')
        self._gen_th('Req.', 'ver-report-steps-hdg-col2')
        self._gen_th('Execution Instructions', 'ver-report-steps-hdg-col3')
        self._gen_th('Expected', 'ver-report-steps-hdg-col4')
        self._gen_th('Actual', 'ver-report-steps-hdg-col5')
        self._gen_th('Pass / Fail', 'ver-report-steps-hdg-col6')
        self._gen_th('Date', 'ver-report-steps-hdg-col7')
        self._gen_tr_close()
        self._gen_thead_close()
        self._gen_tbody_open()

    # -------------------
    ## generates the end of the steps table
    #
    # @returns None
    def _gen_steps_table_close(self):
        self._gen_tbody_close()
        self._gen_table_close()

    # -------------------
    ## generate the overall protocol information
    #
    # @param protocol  the protocol data
    # @return None
    def _gen_protocol_info(self, protocol):
        self._gen_info_table_open()

        # row 0
        start_date = f'<strong>{"Start date": <17}</strong>: '
        if self._do_results:
            start_date += protocol["start_date"]
        self._gen_tr_open(css_class='ver-report-info1')
        self._gen_td(f'<strong>{protocol["proto_id"]}</strong>: {protocol["desc"]}', 'ver-report-info1-col1')
        self._gen_td(start_date, 'ver-report-info1-col2')
        self._gen_tr_close()

        # row 1
        self._gen_tr_open(css_class='ver-report-info1')
        self._gen_td(f'<strong>At</strong>: {protocol["location"]}', 'ver-report-info1-col1')
        self._gen_td('<strong>Requirements</strong> : see below', 'ver-report-info1-col2')
        self._gen_tr_close()

        # row 2
        self._gen_tr_open(css_class='ver-report-info1')
        self._gen_td(f'<strong>Software Version</strong>: {protocol["dut_version"]}', 'ver-report-info1-col1')
        self._gen_td(f'<strong>Serial number</strong>: {protocol["dut_serialno"]}', 'ver-report-info1-col2')
        self._gen_tr_close()

        # row 3
        self._gen_tr_open(css_class='ver-report-info2')
        fmt_str = self._get_list(protocol["objectives"])
        self._fp.write(f'<td colspan="2" class="ver-report-info2-col1"><strong>Objectives</strong>: {fmt_str}</td>\n')
        self._gen_tr_close()

        # row 4
        self._gen_tr_open(css_class='ver-report-info1')
        fmt_str = self._get_list(protocol["preconditions"])
        self._gen_td(f'<strong>Preconditions</strong>: {fmt_str}', 'ver-report-info1-col1')
        fmt_str = self._get_list(protocol["deviations"])
        self._gen_td(f'<strong>Deviations</strong>: {fmt_str}', 'ver-report-info1-col2')
        self._gen_tr_close()

        self._gen_info_table_close()

    # -------------------
    ## converts a list of items into an html unordered list
    # if the list is empty iit generates "N/A" text
    #
    # @param items   the list of items to include in the table
    # @returns None
    def _get_list(self, items):
        if len(items) == 0:
            fmt_list = 'N/A'
        else:
            fmt_list = '<ul>'
            for item in items:
                fmt_list += f'<li>{item}</li>'  # pylint: disable=consider-using-join
            fmt_list += '</ul>'
        return fmt_list

    # -------------------
    ## generates the start of the info table
    #
    # @returns None
    def _gen_info_table_open(self):
        self._gen_table_open('ver-report-table')
        self._gen_tbody_open()

    # -------------------
    ## generates the end of the info table
    #
    # @returns None
    def _gen_info_table_close(self):
        self._gen_tbody_close()
        self._gen_table_close()

    # -------------------
    ## generate title row and one row per step
    #
    # @param steps          the steps data
    # @param requirements   the requirements found in all steps
    # @return None
    def _gen_steps_info(self, steps, requirements):
        self._gen_steps_table_open()
        stepno = 0
        if len(steps) == 0:
            self._gen_tr_open()
            self._gen_td_indented_spanned('No steps found', 7)
            self._gen_tr_close()
        else:
            for step in steps:
                stepno += 1
                self._gen_step(stepno, step, requirements)
        self._gen_steps_table_close()

    # -------------------
    ## generate a step in the test protocol
    #
    # @param stepno         the step number of this step
    # @param step           the step description
    # @param requirements   the one or more reqids for this step
    # @return None
    def _gen_step(self, stepno, step, requirements):
        rs = ResultSummary()
        self._get_overall_result(step, rs)

        self._gen_tr_open()
        self._gen_td(self._get_stepno('min-html', stepno), css_class='ver-report-steps-col1')
        self._gen_td(self._get_reqids('min-html', rs, requirements), css_class='ver-report-steps-col2')
        self._gen_td(self._get_desc('min-html', step), css_class='ver-report-steps-col3')
        self._gen_td(self._get_expected('min-html', rs), css_class='ver-report-steps-col4')
        self._gen_td(self._get_actual('min-html', rs), css_class='ver-report-steps-col5')
        self._gen_td(self._get_result('min-html', rs), css_class='ver-report-steps-col6')
        self._gen_td(self._get_details('min-html', rs, step), css_class='ver-report-steps-col7')
        self._gen_tr_close()

        if self._do_results:
            self._gen_step_comments(step)

    # -------------------
    ## generate the comments for the given step
    #
    # @param step   the step data
    # @return None
    def _gen_step_comments(self, step):
        for comment in step['comments']:
            self._gen_tr_open()
            self._gen_td_indented_spanned(f'<strong>Note:</strong> {comment}', 7)
            self._gen_tr_close()

    # -------------------
    ## generate a requirement table
    #
    # @param requirements   a list of reqmt info
    # @return None
    def _gen_reqmt_table(self, requirements):
        self._gen_blank_line()
        self._gen_reqmts_table_open()
        if len(requirements) == 0:
            self._gen_tr_open()
            self._gen_td('N/A', css_class='ver-report-reqmts-col1')
            self._gen_td('No requirements found', css_class='ver-report-reqmts-col2')
            self._gen_tr_close()
        else:
            for reqmt in sorted(requirements):
                if reqmt in self._requirements:
                    desc = self._requirements[reqmt]['desc']
                else:
                    desc = f'Could not find {reqmt} in file: {services.cfg.reqmt_json_path}'

                self._gen_tr_open()
                self._gen_td(reqmt, css_class='ver-report-reqmts-col1')
                self._gen_td(desc, css_class='ver-report-reqmts-col2')
                self._gen_tr_close()
        self._gen_reqmts_table_close()

    # -------------------
    ## generates the start of the requirements table
    #
    # @returns None
    def _gen_reqmts_table_open(self):
        self._gen_table_open('ver-report-table')
        self._gen_thead_open()
        self._gen_tr_open(css_class='ver-report-hdg1')
        self._gen_th('Req.', 'ver-report-reqmts-hdg-col1')
        self._gen_th('Desc.', 'ver-report-reqmts-hdg-col2')
        self._gen_tr_close()
        self._gen_thead_close()
        self._gen_tbody_open()

    # -------------------
    ## generates the end of the requirements table
    #
    # @returns None
    def _gen_reqmts_table_close(self):
        self._gen_tbody_close()
        self._gen_table_close()

    # -------------------
    ## generate a signature area for all testers
    #
    # @return None
    def _gen_testers(self):
        self._gen_p_strong('Manual Tester Signatures', indent=True)

        self._gen_testers_table_open()
        for initials in sorted(self._testers):
            if initials == 'manual':
                initials = ''
                name = ''
            elif initials in services.cfg.testers:
                name = services.cfg.testers[initials]
            else:
                name = ''

            self._gen_tr_open()
            self._gen_td(initials, css_class='ver-report-testers-col1')
            self._gen_td(name, css_class='ver-report-testers-col2')
            self._gen_td('', css_class='ver-report-testers-col3')
            self._gen_tr_close()

        self._gen_testers_table_close()

    # -------------------
    ## generates the start of the testers table
    #
    # @returns None
    def _gen_testers_table_open(self):
        self._gen_table_open('ver-report-table')
        self._gen_thead_open()
        self._gen_tr_open(css_class='ver-report-hdg1')
        self._gen_th('', 'ver-report-testers-hdg-col1')
        self._gen_th('Tester', 'ver-report-testers-hdg-col2')
        self._gen_th('Signature', 'ver-report-testers-hdg-col3')
        self._gen_tr_close()
        self._gen_thead_close()
        self._gen_tbody_open()

    # -------------------
    ## generates the end of the testers table
    #
    # @returns None
    def _gen_testers_table_close(self):
        self._gen_tbody_close()
        self._gen_table_close()



import os
import sys
import types

import pytest

from medver_pytest import services
from medver_pytest.lib.iuv.iuv_data import IUVData
from medver_pytest.lib.iuv.iuv_requirements import IUVRequirements
from medver_pytest.stdout_logger import StdoutLogger as logger


# -------------------
## Holds content for running an Intended Usa Verification Report
class IUV:
    # -------------------
    ## constructor
    def __init__(self):
        ## holds reference to IUVRequirements object
        self._reqmts = None
        ## holds path to the IUV log file
        self._log_path = None
        ## holds reference to IUVData object
        self._data = None
        ## indicates if an abort() was called
        self._abort_called = False
        ## holds logger lines (if any)
        self._logger_lines = []
        ## holds logger lines (if any)
        self._stdout_lines = []

    # -------------------
    ## return the services object
    #
    # @return the services object
    @property
    def svc(self):
        return services

    # -------------------
    ## return the data object
    #
    # @return the data object
    @property
    def data(self):
        return self._data

    # -------------------
    ## return whether or not abort() was called
    #
    # @return the abort flag
    @property
    def abort_called(self):
        return self._abort_called

    # -------------------
    ## return current list of logger lines
    #
    # @return the logger lines
    @property
    def logger_lines(self):
        return self._logger_lines

    # -------------------
    ## return current list of stdout lines
    #
    # @return the stdout lines
    @property
    def stdout_lines(self):
        return self._stdout_lines

    # -------------------
    ## simulate a print to stdout
    #
    # @param line  the line to save
    # @return None
    def print(self, line):  # print okay
        self._stdout_lines.append(line)

    # -------------------
    ## initialize
    #
    # @return None
    def init(self):
        self._reqmts = IUVRequirements()
        self._reqmts.init()

        self._log_path = os.path.join('out', 'iuv.log')
        self._write('=====')

        self._data = IUVData()
        self._data.init()

        self.init2()

    # -------------------
    ## initialization - stage2
    #
    # @return None
    def init2(self):
        self._abort_called = False
        self._logger_lines = []
        self._stdout_lines = []

    # -------------------
    ## initialize after each test
    #
    # @return None
    def init_each_test(self):
        logger.line('')
        self._stdout_lines = []

    # -------------------
    ## save a logger line
    #
    # @param line   the line that was logged
    # @return None
    def mock_logger(self, line):
        self._logger_lines.append(line)

    # -------------------
    ## clear current logger content
    #
    # @return None
    def clear_logger(self):
        self._logger_lines = []

    # -------------------
    ## indicate abort was called
    #
    # @param msg   optional message to print
    # @return None
    def abort(self, msg=None):
        self._abort_called = True
        logger.line('IUV: iuv.abort() called')
        logger.line(f'IUV: iuv.abort() msg: {msg}')
        logger.line('IUV: exiting testcase (ignore next "SKIPPED")')
        pytest.skip('IUV abort')

    # -------------------
    ## IUV verify - called to verify a value for a particular reqid or set of reqids
    # The requirements are for the IUV report.
    #
    # @param expected   the expected value
    # @param actual     the actual value
    # @param reqids     list of
    # @return None
    def iuv_verify(self, expected, actual, reqids):
        if not services.cfg.iuvmode:
            services.abort('IUV: iuv.iuv_verify() can only be used in --iuvmode')

        tb, location = self._get_location()

        if isinstance(reqids, str):
            reqids = [reqids]
        elif isinstance(reqids, list):
            pass
        else:
            services.logger.err(f'IUV: reqids should be a str or list, found: {type(reqids).__name__}')
            return

        s = []
        for reqid in reqids:
            s.append(reqid.upper())
        reqids = s

        # check reqid was specified
        self._check_reqids(reqids, location)

        expected_type = type(expected).__name__
        actual_type = type(actual).__name__

        # write the info to the IUV log
        if expected == actual:
            res = 'PASS'
            passed = True
        else:
            res = 'FAIL'
            passed = False

        msg = self._write_result(res, location, reqids, expected_type, expected, actual_type, actual)

        # save the reqid to the trace file
        self._reqmts.save_reqids_to_trace(reqids, passed, location)

        if not passed:
            raise AssertionError(msg).with_traceback(tb)

    # -------------------
    ## get the location of the calling iuv_verify() function
    #
    # @return None
    def _get_location(self):
        tb = None
        try:
            raise AssertionError
        except AssertionError:
            tb = sys.exc_info()[2]

        # go two caller back
        frame = tb.tb_frame
        frame = frame.f_back
        frame = frame.f_back

        tb = types.TracebackType(tb_next=None,
                                 tb_frame=frame,
                                 tb_lasti=frame.f_lasti,
                                 tb_lineno=frame.f_lineno)

        location = f'{tb.tb_frame.f_code.co_filename}({tb.tb_frame.f_lineno})'
        return tb, location

    # -------------------
    ## check the reqmt ids passed in are valid and recognized values for the IUV SRS
    #
    # @param reqids    the reqid list for this call
    # @param location  the iuv_verify() location
    # @return None
    def _check_reqids(self, reqids, location):
        for reqid in reqids:
            msg = self._reqmts.check(reqid, location)
            if msg is not None:
                services.abort(msg)

    # -------------------
    ## write the result of the iuv_verify call to the log and to the trace json file
    #
    # @param result          either PASS or FAIL
    # @param location        the iuv_verify() location
    # @param reqids          the reqid list for this call
    # @param expected_type   the type of the expected value
    # @param expected        the expected value
    # @param actual_type     the type of the actual value
    # @param actual          the actual value
    # @return the formatted message line
    def _write_result(self, result, location, reqids, expected_type, expected, actual_type, actual):
        msg = f'{result} verify: {location}\n' \
              f'     Expected: ({expected_type: <5}): {expected}\n' \
              f'     Actual  : ({actual_type: <5}): {actual}'
        self._write(msg)

        # get the requirement desc
        for reqid in reqids:
            desc = self._reqmts.get(reqid)
            self._write(f'     Reqid   : "{reqid}" {desc}')

        self._write('')
        return msg

    # -------------------
    ## write the given line to the log; write the trace json file to disk
    #
    # @param line  the line to write to the IUV log
    # @return None
    def _write(self, line):
        with open(self._log_path, 'a', encoding='utf-8') as log:
            log.write(f'{line}\n')

        self._reqmts.save_trace()

    # -------------------
    ## print report summary of IUV PASS/FAIL to stdout
    #
    # @return None
    def report(self):
        logger.line('')

        ## see function self._log
        self._reqmts.report_trace(self._log)
        self._reqmts.report_missing(self._log)

    # --------------------
    ## print a line to stdout
    #
    # @param tag    a tag for logging
    # @param val    the value to write
    # @param total  optional total value used to calculate a percentage
    # @return None
    def _log(self, tag, val=None, total=None):
        if tag == '':
            logger.line('')
            return

        if val is None:
            logger.line(tag)
        elif total is None:
            logger.line(f'   {tag: <25} {val: >5}')
        elif total == 0:
            logger.line(f'   {tag: <25} {val: >5} (total is 0)')
        else:
            logger.line(f'   {tag: <25} {val: >5} {round(val / total * 100.0, 0):>5}%')



from reportlab.lib import colors
from reportlab.lib.units import inch
from reportlab.platypus import ListFlowable
from reportlab.platypus import ListItem
from reportlab.platypus import Paragraph
from reportlab.platypus import Table
from reportlab.platypus import TableStyle

from .gen_tp_base import GenTpBase
from ..common.gen_base_pdf import GenBasePdf
from medver_pytest import services
from medver_pytest.result_summary import ResultSummary


# -------------------
## Generates a Test Protocol report in PDF format
class GenTpPdf(GenBasePdf, GenTpBase):
    # -------------------
    ## constructor
    #
    # @param protocols   the data to use
    # @param do_results  generate results or not
    def __init__(self, protocols, do_results=True):
        ## see base clase for doc
        self._doc_path = None
        ## see base clase for doc
        self._testers = None

        GenBasePdf.__init__(self)
        GenTpBase.__init__(self, protocols, do_results)

        self._init('pdf', protocols, do_results)
        # uncomment to debug:
        # from medver_pytest.lib.stdout_logger import StdoutLogger as logger
        # logger.dbg(f'protocols: {json.dumps(self._protocols, indent=2)}')

        self._doc_init(self._doc_path)

        # === styles
        ## holds reference to the style sheet for most text in the doc
        self._stbody = self._style_sheet['BodyText']
        ## holds reference to the style sheet for text in a table row
        self._strow = self._style_sheet['BodyText']
        ## holds reference to the style sheet for text in monospace font
        self._stmono = self._style_sheet['Mono']

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()

        self._gen_test_run_details()
        title = self._get_title()
        self._gen_title(title)

        for _, protocol in self._protocols.items():
            self._gen_protocol(protocol)

        if self._testers:
            self._gen_testers()

        self._build()

    # -------------------
    ## generate a protocol table using the given data
    #
    # @param protocol   the data to use
    # @return None
    def _gen_protocol(self, protocol):
        tbl = []
        self._gen_protocol_info(tbl, protocol)

        requirements = []
        need_space = self._gen_steps_info(tbl, protocol['steps'], requirements)

        self._gen_protocol_table(tbl, need_space)
        self._gen_reqmt_table(requirements)

    # -------------------
    ## generate protocol info
    #
    # @param tbl        the table to write to
    # @param protocol   the data to use
    # @return None
    def _gen_protocol_info(self, tbl, protocol):
        if self._do_results:
            start_date = f'<b>Start date:</b>: {protocol["start_date"]}'
        else:
            start_date = '<b>Start date:</b>: '

        # row 0
        tbl.append([
            Paragraph(f'<b>{protocol["proto_id"]}</b>: {protocol["desc"]}', self._stbody),
            '', '', '',
            Paragraph(start_date, self._stbody),
            '', '',
        ])

        # row 1
        tbl.append([
            Paragraph(f'<b>At</b>: {protocol["location"]}', self._stbody),
            '', '', '',
            Paragraph('<b>Requirements:</b>: see below', self._stbody),
            '', '',
        ])

        # row 2
        tbl.append([
            Paragraph(f'<b>Software Version</b>: {protocol["dut_version"]}', self._stbody),
            '', '', '',
            Paragraph(f'<b>Serial number:</b> {protocol["dut_serialno"]}', self._stbody),
            '', '',
        ])

        # row 3
        tbl.append([
            self._gen_list('Objectives', protocol['objectives']),
            '', '', '', '', '',
        ])

        # row 4
        tbl.append([
            self._gen_list('Preconditions', protocol['preconditions']),
            '', '', '',
            self._gen_list('Deviations', protocol['deviations']),
            '', '',
        ])

    # -------------------
    ## generate a list using the given items
    #
    # @param tag      the name of the list
    # @param items    the list of data to use
    # @return a paragraph containing the formatted list
    def _gen_list(self, tag, items):
        paragraph_text = f'<b>{tag}</b>'
        if len(items) == 0:
            list_items = 'N/A'
            paragraph = Paragraph(f'{paragraph_text}: {list_items}', self._stbody)
        elif len(items) == 1:
            list_items = items[0]
            paragraph = Paragraph(f'{paragraph_text}: {list_items}', self._stbody)
        else:
            # more than 1 item in the list
            paragraph = [
                Paragraph(f'{paragraph_text}:', self._stbody),
            ]
            list_items = []
            for item in items:
                list_items.append(ListItem(Paragraph(item)))
            paragraph.append(ListFlowable(list_items, bulletType='bullet'))

        return paragraph

    # -------------------
    ## generate title row and one row per step
    #
    # @param tbl            the table to write to
    # @param steps          the steps data
    # @param requirements   the requirements found in all steps
    # @return None
    def _gen_steps_info(self, tbl, steps, requirements):
        # row 5: title row for step info
        tbl.append(
            [
                Paragraph('<b>Step</b>', self._stbody),
                Paragraph('<b>Req.</b>', self._stbody),
                Paragraph('<b>Execution Instructions</b>', self._stbody),
                Paragraph('<b>Expected</b>', self._stbody),
                Paragraph('<b>Actual</b>', self._stbody),
                Paragraph('<b>Pass /\nFail</b>', self._stbody),
                Paragraph('<b>Date</b>', self._stbody)
            ]
        )

        need_space = False
        if len(steps) == 0:
            tbl.append([Paragraph('N/A', self._strow), '', Paragraph('No steps found', self._strow), '', '', '', ''])
        else:
            # row N: one row for each step
            stepno = 0
            for step in steps:
                stepno += 1
                ns = self._gen_step(tbl, stepno, step, requirements)
                need_space = need_space or ns

        return need_space

    # -------------------
    ## generate a step in the test protocol
    #
    # @param tbl            the table to generate into
    # @param stepno         the step number of this step
    # @param step           the step description
    # @param requirements   the one or more reqids for this step
    # @return a paragraph containing the formatted list
    def _gen_step(self, tbl, stepno, step, requirements):
        # default is a passed, and empty result
        rs = ResultSummary()
        self._get_overall_result(step, rs)

        stepno = Paragraph(self._get_stepno('pdf', stepno), self._strow)
        reqids_str = Paragraph(self._get_reqids('pdf', rs, requirements), self._strow)
        desc = Paragraph(self._get_desc('pdf', step), self._strow)
        expected = Paragraph(self._get_expected('pdf', rs), self._stmono)
        actual = Paragraph(self._get_actual('pdf', rs), self._stmono)
        result = Paragraph(self._get_result('pdf', rs), self._strow)
        details = Paragraph(self._get_details('pdf', rs, step), self._strow)
        tbl.append([stepno, reqids_str, desc, expected, actual, result, details])

        if self._do_results:
            self._gen_step_comments(tbl, step)

        # if the actual or expected is long-ish, indicate those columsn will need more space
        if services.cfg.page_info.orientation == 'portrait':
            max_len = 10
        else:
            max_len = 20
        if len(str(rs.actual_formatted)) > max_len:
            return True
        if len(str(rs.expected_formatted)) > max_len:
            return True

        return False

    # -------------------
    ## generate the comments for the given step
    #
    # @param tbl    the table to add the comments to
    # @param step   the step data
    # @return None
    def _gen_step_comments(self, tbl, step):
        for comment in step['comments']:
            para = Paragraph(f'Note: {comment}', self._strow)
            tbl.append(['', '', para, '', '', '', ''])

    # -------------------
    ## generate a protocol table
    #
    # @param tbl          the table to draw into
    # @param need_space   indicates if the table requires a different format because of space layout
    # @return None
    def _gen_protocol_table(self, tbl, need_space):
        # these widths are the same portrait, landscape, needs space or not
        stepno_width = 0.41 * inch
        reqmt_width = 0.65 * inch
        details_width = 1.47 * inch
        result_width = 0.47 * inch
        if services.cfg.page_info.orientation == 'portrait':
            # portrait: should be 7.5"
            if need_space:
                value_width = 1.31 * inch
                tbl_widths = [
                    stepno_width,  # stepno
                    reqmt_width,  # reqmt.
                    1.88 * inch,  # desc
                    value_width,  # expected
                    value_width,  # actual
                    result_width,  # result
                    details_width,  # details
                ]
            else:
                value_width = 0.70 * inch
                tbl_widths = [
                    stepno_width,  # stepno
                    reqmt_width,  # reqmt.
                    3.10 * inch,  # desc
                    value_width,  # expected
                    value_width,  # actual
                    result_width,  # result
                    details_width,  # details
                ]
        else:
            # landscape (10" width)
            if need_space:
                value_width = 1.68 * inch
                tbl_widths = [
                    stepno_width,  # stepno
                    reqmt_width,  # reqmt.
                    3.64 * inch,  # desc
                    value_width,  # expected
                    value_width,  # actual
                    result_width,  # result
                    details_width,  # details
                ]
            else:
                value_width = 1.00 * inch
                tbl_widths = [
                    stepno_width,  # stepno
                    reqmt_width,  # reqmt.
                    5.00 * inch,  # desc
                    value_width,  # expected
                    value_width,  # actual
                    result_width,  # result
                    details_width,  # details
                ]

        # uncomment to debug
        # total = 0.0
        # for val in tbl_widths:
        #     total += val
        # # there are 72 per inch
        # logger.dbg(f"{total / 72: >8.2f} need_space:{need_space}")

        t = Table(tbl,
                  colWidths=tbl_widths,
                  spaceBefore=0.25 * inch,
                  spaceAfter=0.0 * inch,
                  )

        t.setStyle(TableStyle(
            [
                # borders
                ('GRID', (0, 0), (-1, -1), 0.25, colors.black),

                # row 0: spans 2 cols, tp info and desc
                ('BACKGROUND', (0, 0), (-1, 0), colors.lightblue),
                ('SPAN', (0, 0), (4, 0)),
                ('SPAN', (4, 0), (-1, 0)),
                ('INNERGRID', (0, 0), (3, 0), 0.25, colors.black),

                # row 1: spans 2 cols, executed_by and requirements
                ('BACKGROUND', (0, 1), (-1, 1), colors.lightblue),
                ('SPAN', (0, 1), (4, 1)),
                ('SPAN', (4, 1), (-1, 1)),

                # row 2: spans 2 cols, version and serialno
                ('BACKGROUND', (0, 2), (-1, 2), colors.lightblue),
                ('SPAN', (0, 2), (4, 2)),
                ('SPAN', (4, 2), (-1, 2)),

                # row 3: spans all cols, objectives
                ('BACKGROUND', (0, 3), (-1, 3), colors.lightblue),
                ('SPAN', (0, 3), (-1, 3)),

                # row 4: spans 2 cols, preconditions and deviations
                ('BACKGROUND', (0, 4), (-1, 4), colors.lightblue),
                ('SPAN', (0, 4), (4, 4)),
                ('SPAN', (4, 4), (-1, 4)),

                # use borders from now on
                # row 5: title row with all columns
                ('BACKGROUND', (0, 5), (-1, 5), colors.lightgrey),

                # row 6 and on
                ('VALIGN', (0, 6), (0, -1), 'MIDDLE'),  # 0: stepno ; middle/center
                ('ALIGN', (0, 6), (0, -1), 'CENTER'),
                ('VALIGN', (1, 6), (1, -1), 'MIDDLE'),  # 1: reqmt
                ('VALIGN', (2, 6), (2, -1), 'MIDDLE'),  # 2: desc
                ('VALIGN', (3, 6), (3, -1), 'MIDDLE'),  # 3: expected
                ('VALIGN', (4, 6), (4, -1), 'MIDDLE'),  # 4: actual
                ('VALIGN', (5, 6), (5, -1), 'MIDDLE'),  # 5: result ; middle/center
                ('ALIGN', (5, 6), (5, -1), 'CENTER'),
                ('VALIGN', (6, 6), (6, -1), 'MIDDLE'),  # details
            ]
        ))
        self._elements.append(t)

    # -------------------
    ## generate a requirement table
    #
    # @param requirements   a list of reqmt info
    # @return None
    def _gen_reqmt_table(self, requirements):
        if services.cfg.page_info.orientation == 'portrait':
            # portrait: 7.5"
            tbl_widths = [
                0.85 * inch,  # reqmt.
                6.65 * inch,  # desc
            ]
        else:
            # landscape: 10"
            tbl_widths = [
                0.85 * inch,  # reqmt.
                9.15 * inch,  # desc
            ]

        # row 0: title row
        tbl = [
            [
                Paragraph('<b>Req.</b>', self._strow),
                Paragraph('<b>Desc.</b>', self._strow),
            ]
        ]

        if len(requirements) == 0:
            tbl.append([Paragraph('N/A', self._strow), Paragraph('No requirements found', self._strow)])
        else:
            for reqmt in sorted(requirements):
                reqmt_str = Paragraph(reqmt, self._strow)
                if reqmt in self._requirements:
                    desc = self._requirements[reqmt]['desc']
                else:
                    desc = f'Could not find {reqmt} in file: {services.cfg.reqmt_json_path}'
                tbl.append([reqmt_str, Paragraph(desc, self._strow)])

        t = Table(tbl,
                  colWidths=tbl_widths,
                  spaceBefore=0.1 * inch,
                  spaceAfter=0.50 * inch,
                  )

        t.setStyle(TableStyle(
            [
                # borders
                ('GRID', (0, 0), (-1, -1), 0.25, colors.black),

                # use borders from now on
                # row 0: title row with all columns
                ('BACKGROUND', (0, 0), (-1, 0), colors.lightgrey),

                # row 1 and on
                ('VALIGN', (0, 1), (0, -1), 'MIDDLE'),  # col 0: reqmt
                ('VALIGN', (1, 1), (1, -1), 'MIDDLE'),  # col 1: desc
            ]
        ))

        self._elements.append(t)

    # -------------------
    ## generate a tester signature area
    #
    # @return None
    def _gen_testers(self):
        para = Paragraph('<b>Manual Tester Signatures</b>', self._style_sheet['BodyText'])
        self._elements.append(para)

        if services.cfg.page_info.orientation == 'portrait':
            # portrait: 7.5"
            tbl_widths = [
                0.50 * inch,  # initials
                3.50 * inch,  # name
                3.50 * inch,  # signature
            ]
        else:
            # landscape: 10"
            tbl_widths = [
                0.50 * inch,  # initials
                3.25 * inch,  # name
                6.25 * inch,  # signature
            ]

        # row 0: title row
        tbl = [
            [
                Paragraph('<b></b>', self._strow),
                Paragraph('<b>Tester</b>', self._strow),
                Paragraph('<b>Signature</b>', self._strow),
            ]
        ]

        for initials in sorted(self._testers):
            if initials == 'manual':
                initials = ''
                name = ''
            elif initials in services.cfg.testers:
                name = services.cfg.testers[initials]
            else:
                name = ''
            tbl.append([Paragraph(initials, self._strow),
                        Paragraph(name, self._strow),
                        Paragraph('\n\n', self._strow)
                        ])

        t = Table(tbl,
                  colWidths=tbl_widths,
                  spaceBefore=0.1 * inch,
                  spaceAfter=0.50 * inch,
                  )

        t.setStyle(TableStyle(
            [
                # borders
                ('GRID', (0, 0), (-1, -1), 0.25, colors.black),

                # use borders from now on
                # row 0: title row with all columns
                ('BACKGROUND', (0, 0), (-1, 0), colors.lightgrey),

                # row 1 and on
                ('VALIGN', (0, 1), (0, -1), 'MIDDLE'),  # col 0: initials
                ('VALIGN', (1, 1), (1, -1), 'MIDDLE'),  # col 1: name
                ('VALIGN', (1, 1), (1, -1), 'MIDDLE'),  # col 1: signature
            ]
        ))

        self._elements.append(t)



import tkinter as tk

from .dlg_manual_action import DlgManualAction
from .dlg_manual_verify import DlgManualVerify


# -------------------
## holds functions for displaying a GUI component for manual interactions
class Gui:
    # -------------------
    ## constructor
    def __init__(self):
        # create base window

        ## holds the base window
        self._root = tk.Tk()

        # disable 'x' button'
        self._root.overrideredirect(True)

        # make it invisible
        self._root.withdraw()

    # -------------------
    ## initialization
    #
    # @return None
    def init(self):
        pass

    # -------------------
    ## terminate
    #
    # @return None
    def term(self):
        pass

    # -------------------
    ## display a dialog box for a manual action to be performed by the tester
    #
    # @param proto_id   the test protocol id
    # @param action     the action for the tester to perform
    # @return None
    def show_step_action(self, proto_id, action):
        dlg = DlgManualAction()
        state, reason = dlg.step_manual_action(self._root,
                                               f'{proto_id}: perform manual action',
                                               action)
        return state, reason

    # -------------------
    ## display a dialog box for a manual action to be performed by the tester
    #
    # @param proto_id     the test protocol id
    # @param verify_desc  the verification for the tester to perform
    # @param expected     the expected value for the verification
    # @return None
    def show_verify(self, proto_id, verify_desc, expected):
        dlg = DlgManualVerify()
        state, actual = dlg.manual_verify(self._root,
                                          f'{proto_id}: perform manual verification',
                                          verify_desc,
                                          expected)
        return state, actual



import argparse

from medver_pytest import services


# -------------------
## handles CLI interface via conftest.py and normal argparse
class Cli:
    # -------------------
    ## initialize
    def __init__(self):
        ## holds true/false for running in IUV mode
        self.iuvmode = 'none'
        ## holds path to cfg.json file
        self.cfg_path = 'none'
        ## holds optional test run type; overrides cfg.json value
        self.test_run_type = 'none'
        ## holds optional test run id; overrides cfg.json value
        self.test_run_id = 'none'

    # -------------------
    ## conftest.py: add pytest options
    #
    # @param parser  the conftest parser object
    # @return None
    def addoption(self, parser):
        # note: use 'none' since default=None sets it to "None" not None
        parser.addoption('--iuvmode', action='store_true', dest='iuvmode', default='none')
        parser.addoption('--cfg_path', action='store', dest='cfg_path', default='none')
        parser.addoption('--test_run_type', action='store', dest='test_run_type', default='none')
        parser.addoption('--test_run_id', action='store', dest='test_run_id', default='none')

    # -------------------
    ## conftest.py: handle CLI overrides
    #
    # @param config  the conftest config object
    # @return None
    def configure(self, config):
        self.iuvmode = config.getoption('iuvmode')
        self.cfg_path = str(config.getoption('cfg_path'))

        self.test_run_type = config.getoption('test_run_type')
        self.test_run_id = config.getoption('test_run_id')

    # -------------------
    ## argparse: add cli arguments
    #
    # @return None
    def parse(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--iuvmode', action='store_true', dest='iuvmode', default='none')
        parser.add_argument('--cfg_path', action='store', dest='cfg_path', default='none')
        parser.add_argument('--test_run_type', action='store', dest='test_run_type', default='none')
        parser.add_argument('--test_run_id', action='store', dest='test_run_id', default='none')
        args, _ = parser.parse_known_args()

        self.iuvmode = args.iuvmode
        self.cfg_path = args.cfg_path
        self.test_run_type = args.test_run_type
        self.test_run_id = args.test_run_id

    # -------------------
    ## set values that are set i.e. non-overridable.
    # These are set before the cfg json file is read.
    #  * cfg_path
    #  * iuvmode
    #
    # @return None
    def save(self):
        # should only have non-overridable values in here
        if self.cfg_path != 'none':
            services.cfg.cfg_path = self.cfg_path
        if self.iuvmode != 'none':
            services.cfg.iuvmode = self.iuvmode

    # -------------------
    ## set values that override existing cfg values
    #  * test_run_type
    #  * test_run_id
    #
    # @return None
    def override(self):
        if self.test_run_type != 'none':
            services.cfg.test_run_type = self.test_run_type
        if self.test_run_id != 'none':
            services.cfg.test_run_id = self.test_run_id

    # -------------------
    ## check if a switch name is valid
    #
    # @param name   the switch name to check
    # @return True if valid, False otherwise
    def is_valid(self, name):
        return name in ['cfg_path', 'iuvmode', 'test_run_type', 'test_run_id']



import os

from medver_pytest import services
from medver_pytest.constants import Constants
from medver_pytest.requirements import Requirements


# -------------------
## Common functions for TP reports
class GenTpBase:  # pylint: disable=too-few-public-methods
    # -------------------
    ## constructor
    #
    # @param protocols   the data to use
    # @param do_results  generate results or not
    def __init__(self, protocols, do_results):
        ## holds protocol inforation
        self._protocols = protocols
        ## holds flag to generate results (for TP report), or not (for TP doc)
        self._do_results = do_results

        ## holds requirements information
        self._requirements = []
        ## the path and file name of the generated document
        self._doc_path = None
        ## the extension of the document's file name
        self._doc_extension = None

        ## holds list of testers
        self._testers = []

    # -------------------
    ## initialization
    #
    # @param extension     the file name extension for the report file
    # @param protocols     the list of protocol objects for the report
    # @param do_results    flag indicating if it is (False) a test protocol or (True) a test report
    # @return None
    def _init(self, extension, protocols, do_results):
        ## holds protocol inforation
        self._protocols = protocols
        ## holds flag to generate results (for TP report), or not (for TP doc)
        self._do_results = do_results

        self._doc_extension = extension
        self._set_doc_path()

    # -------------------
    ## full document path
    #
    # @return None
    def _set_doc_path(self):
        if self._do_results:
            fname = Constants.tp_report_fname
        else:
            fname = Constants.tp_protocol_fname

        # hack
        if self._doc_extension == 'min-html':
            self._doc_path = os.path.join(services.cfg.outdir, f'{fname}-min.html')
        else:
            self._doc_path = os.path.join(services.cfg.outdir, f'{fname}.{self._doc_extension}')

    # -------------------
    ## initialization for generating the doc
    #
    # @return None
    def _gen_init(self):
        if self._do_results:
            services.logger.start(f'report: TP with results ({self._doc_extension})')
        else:
            services.logger.start(f'report: TP without results ({self._doc_extension})')

        self._requirements = Requirements.fast_load()

    # -------------------
    ## get the document title
    #
    # @return the title
    def _get_title(self):
        if self._do_results:
            title = 'Test Protocols with results'
        else:
            title = 'Test Protocols'
        return title

    # -------------------
    ## set the overall result (pass/fail) in the given result summary
    #
    # @param step   the current step object
    # @param rs     the current result summary object
    # @return None
    def _get_overall_result(self, step, rs):
        rs.passed()

        # start as passed, load each result
        # if they all passed, then display the last one
        # if any failed, then stop and display the first failed one
        # in all cases, the list of unique reqids in all results
        for res in step['results']:
            rs.append_result(res)
            if rs.result == 'FAIL':
                break

    # -------------------
    ## get the stepno ready to print for the given document type
    #
    # @param doctype  the current document type
    # @param stepno   the current stepno
    # @return the formatted string
    def _get_stepno(self, doctype, stepno):  # pylint: disable=unused-argument
        return str(stepno)

    # -------------------
    ## get the step description ready to print for the given document type
    #
    # @param doctype  the current document type
    # @param step     the current step
    # @return the formatted string
    def _get_desc(self, doctype, step):  # pylint: disable=unused-argument
        return str(step['desc'])

    # -------------------
    ## get the reqmt ids to print for the given document type
    #
    # @param doctype       the current document type
    # @param rs            the current result summary
    # @param requirements  a list of requirement ids found
    # @return the formatted string
    def _get_reqids(self, doctype, rs, requirements):
        # docx, txt
        if doctype in ['pdf', 'min-html']:
            delim = '<br/>'
        else:
            delim = '\n'

        if rs.reqids is None or rs.reqids == {}:
            reqids_str = 'N/A'
        else:
            for reqid in rs.reqids:
                if reqid not in requirements:
                    requirements.append(reqid)
            sorted_reqids = sorted(rs.reqids.keys())
            if len(sorted_reqids) == 1:
                reqids_str = sorted_reqids[0]
            else:
                reqids_str = delim.join(sorted_reqids)
        return reqids_str

    # -------------------
    ## get the expected value ready to print for the given document type
    #
    # @param doctype  the current document type
    # @param rs       the result summary object
    # @return the formatted string
    def _get_expected(self, doctype, rs):
        if doctype in ['txt', 'docx']:
            val = str(rs.expected_formatted)
        else:
            # pdf, min-html: can have embedded html; to convert
            val = str(rs.expected_formatted).replace('<', '&lt;')
            val = val.replace('>', '&gt;')
        return val

    # -------------------
    ## get the actual value ready to print for the given document type
    #
    # @param doctype  the current document type
    # @param rs       the result summary object
    # @return the formatted string
    def _get_actual(self, doctype, rs):
        val = ''
        if self._do_results:
            if doctype in ['txt', 'docx']:
                val = str(rs.actual_formatted)
            else:
                # pdf, min-html: can have embedded html; to convert
                val = str(rs.actual_formatted).replace('<', '&lt;')
                val = val.replace('>', '&gt;')

        return val

    # -------------------
    ## get the result (PASS/FAIL) ready to print for the given document type
    #
    # @param doctype  the current document type
    # @param rs       the result summary object
    # @return the formatted string
    def _get_result(self, doctype, rs):  # pylint: disable=unused-argument
        val = ''
        if self._do_results:
            val = str(rs.result)
        return val

    # -------------------
    ## get the step details ready to print for the given document type
    #
    # @param doctype  the current document type
    # @param rs       the result summary object
    # @param step     the current step object
    # @return the formatted string
    def _get_details(self, doctype, rs, step):
        if doctype == 'docx':
            delim = '\n'
        elif doctype in ['pdf', 'min-html']:
            delim = '<br/>'
        else:
            delim = ', '

        if self._do_results:
            val = str(step['dts'])
        else:
            val = ''

        tester = step['tester']
        val += delim + 'tester: ' + tester
        if tester != 'auto':
            if tester not in self._testers:
                self._testers.append(tester)
        val += delim + str(rs.location)

        return val



import glob
import os
import shutil

from .storage_file_base import StorageFileBase
from medver_pytest import services


# -------------------
## Holds storage information and how/where to save it
class StorageSharedDrive(StorageFileBase):
    # -------------------
    ## constructor
    def __init__(self):
        ## holds the path to the protocol json file
        super().__init__()

    # -------------------
    ## initialize
    #
    # @return None
    def init(self):
        # check storage_shared_path is defined
        if services.cfg.storage_shared_dir is None:
            services.abort('storage_shared_dir is None')

        # check if the dir exists
        if not os.path.isdir(services.cfg.storage_shared_dir):
            services.abort(f'storage_shared_dir does not exist: {services.cfg.storage_shared_dir}')

        services.logger.ok(f'found shared directory: {services.cfg.storage_shared_dir}')

    # -------------------
    ## terminate
    #
    # @return None
    def term(self):
        if services.cfg.report_mode:
            return

        services.logger.ok(f'publishing to shared: {services.cfg.storage_shared_dir}')
        self._save_to_shared_folder('*_trace.json')
        self._save_to_shared_folder('*_summary.json')
        self._save_to_shared_folder('*_protocol.json')

    # -------------------
    ## copy generated files to the shared directory
    #
    # @param pattern  the pattern of files to match usually json files
    # @return None
    def _save_to_shared_folder(self, pattern):
        for f in glob.glob(os.path.join(services.cfg.outdir, pattern)):
            shutil.copy2(f, services.cfg.storage_shared_dir)



import tkinter as tk

from medver_pytest import services
from medver_pytest.os_specific import OsSpecific


# -------------------
## holds base class for all manual dialog boxes
class DlgBase:  # pylint: disable=too-few-public-methods
    ## constant: width of the dialog box
    width = 850
    ## constant: height of the dialog box
    height = 650
    ## constant: holds font to use for btns and most text and entry widgets
    common_font = ('TkDefaultFont', 13)
    ## constant: holds font to use for smaller areas
    small_font = ('TkDefaultFont', 10, 'bold')
    ## constant: width of the first column for buttons
    col0_btn_width = 11  # characters
    ## constant: width of the first column for text widgets
    col0_txt_width = 14  # characters
    ## constant: width of col 1 and 2
    col12_width = 27

    # -------------------
    ## constructor
    def __init__(self):
        ## main TK window
        self._parent = None

        ## dialog box object
        self._dlg = None
        ## dialog frame
        self._frame = None

        # tester info
        ## the testers name if found in cfg
        self._tester_name = 'unknown'
        ## the tester initials text box
        self._txt_tester_initials = None
        ## the tester name lbl
        self._lbl_tester_name = None

        ## the Next button
        self._btn_next = None

        ## the Message label
        self._lbl_message = None

        ## callback to do additional actions when the initials widget is changed
        self._initials_key_pressed_callback = None

    # -------------------
    ## common first portion of the dlg initialization
    #
    # @param row        the row to use
    # @param msg        the initial message for the msg textbox
    # @param title      the dlg title to use
    # @param fn_ensure_initials  the callback function to use when the initials are set/changed
    # @return last row number
    def _dlg_common_first(self, row, title, msg, fn_ensure_initials):
        services.logger.user(f'display dlg: {title}')
        self._create_dlg(title)

        self._create_message_box(row, msg)

        row += 1

        # disable buttons if initials is empty
        self._initials_key_pressed_callback = fn_ensure_initials

        self._create_tester_info(row)

        return row

    # -------------------
    ## common last portion of the dlg initialization
    #
    # @param row        the row to use
    # @param fn_handle_click     the callback function to use when a click occurs
    # @param fn_ensure_initials  the callback function to use when the initials are set/changed
    # @return None
    def _dlg_common_last(self, row, fn_handle_click, fn_ensure_initials):
        row += 1
        self._create_empty_row(row)

        row += 1
        self._create_next_btn(row, 0, lambda: fn_handle_click('next'))
        self._create_abort_btn(row, 1, lambda: fn_handle_click('abort'))

        row += 1
        self._create_empty_row(row)

        fn_ensure_initials()
        self._set_dlg_location()

    # -------------------
    ## create the TopLevel dialog box
    #
    # @param title   the name of the dlg box
    # @return None
    def _create_dlg(self, title):
        self._dlg = tk.Toplevel(self._parent)

        # temporarily make iconify it (make it invisible)
        self._dlg.withdraw()

        self._dlg.title(title)

        self._dlg.resizable(width=tk.FALSE, height=tk.FALSE)  # remove the max window button
        ## set callback for Delete Window
        self._dlg.protocol('WM_DELETE_WINDOW', self._disable_event)

        self._dlg.config(bg='lightgrey')
        self._create_frame()

    # -------------------
    ## set the dlg location based on the last position saved, if any.
    # otherwise center the dlg in the current screen
    #
    # @return None
    def _set_dlg_location(self):
        # dlgbox is created, but all content may not be updated yet
        self._parent.update_idletasks()
        self._dlg.update_idletasks()

        # show the window
        self._dlg.deiconify()

        # use last position if non-0, otherwise center it
        if services.status.last_position_saved():
            OsSpecific.init()
            if OsSpecific.os_name == 'macos':
                offset = 0
            else:
                # note: the 37 is the height of the OS provided top bar
                offset = 37
            self._dlg.geometry(f'+{services.status.last_position_x}+{services.status.last_position_y - offset}')
            self._dlg.update_idletasks()
        else:
            # center the window
            self._parent.eval(f'tk::PlaceWindow {str(self._dlg)} center')
            self._dlg.update_idletasks()
            services.status.save_last_position(self._dlg.winfo_x(), self._dlg.winfo_y())

        ## set callback for Configure
        self._dlg.bind('<Configure>', self._on_move)

    # -------------------
    ## callback when the dlg is moved
    #
    # @param event   the tkinter config event that occurred
    # @return None
    def _on_move(self, event):  # pylint: disable=unused-argument
        # note: this will get events for all children within the dlg box

        # uncomment to debug
        # services.logger.dbg(f'move: {self._dlg.winfo_x()}, {self._dlg.winfo_y()}')

        # save the dlg box frame x,y
        services.status.save_last_position(self._dlg.winfo_x(), self._dlg.winfo_y())

    # -------------------
    ## used to ignore click events
    #
    # @return None
    def _disable_event(self):
        pass

    # -------------------
    ## create the Frame that holds the grid of buttons, text, etc.
    #
    # @return None
    def _create_frame(self):
        self._frame = tk.Frame(self._dlg,
                               bg='lightgrey')
        self._frame.pack(padx=1, pady=5)
        self._frame.grid_columnconfigure(0, weight=1)

    # -------------------
    ## create the tester information: initials and name.
    # the name is retried from the cfg file based on the given initials.
    # there can be 2 or 3 initials only
    #
    # @param row   the row to place these widgets in
    # @return None
    def _create_tester_info(self, row):
        self._create_tester_initials(row, 0)
        self._create_tester_name(row, 1)

    # -------------------
    ## create the tester initials Entry box
    #
    # @param row   the row to place the widget in
    # @param col   the column to place the widget in
    # @return None
    def _create_tester_initials(self, row, col):
        self._txt_tester_initials = tk.Entry(self._frame,
                                             width=self.col0_txt_width,
                                             font=self.common_font,
                                             bg='lightgrey',
                                             fg='black',
                                             highlightbackground='black', highlightthickness=1
                                             )
        services.logger.user(f'tester initials: "{services.status.last_tester_initials}"')
        self._txt_tester_initials.insert(0, services.status.last_tester_initials)
        self._txt_tester_initials.focus()
        ## set callback for KeyRelease
        self._txt_tester_initials.bind('<KeyRelease>', self._initials_key_pressed)
        self._txt_tester_initials.grid(row=row, column=col, sticky='W')

    # -------------------
    ## indicates if the tester initials has 2 or 3 characters
    #
    # @return True if 2 or 3 chars, False otherwise
    def _tester_initials_is_wrong(self):
        num = len(services.status.last_tester_initials)
        return num < 2 or num > 3

    # -------------------
    ## create tester name label. Filled based on tester initials and content of cfg file
    #
    # @param row  the row to place it in
    # @param col  the column to place it in
    # @return None
    def _create_tester_name(self, row, col):
        self._get_tester_name()
        self._lbl_tester_name = tk.Label(self._frame,
                                         text=self._tester_name,
                                         anchor='w',
                                         width=self.col12_width,
                                         height=1,
                                         font=self.common_font,
                                         fg='black',
                                         bg='lightgrey')
        self._lbl_tester_name.grid(row=row, column=col, sticky='NEWS', columnspan=2, ipady=5)

    # -------------------
    ## handle an entry in the tester initial box
    #
    # @param unused   ignored; the key that was pressed
    # @return None
    # noinspection PyUnusedLocal
    def _initials_key_pressed(self, unused):  # pylint: disable=unused-argument
        services.status.last_tester_initials = self._txt_tester_initials.get().upper().strip()
        self._get_tester_name()
        self._lbl_tester_name.config(text=self._tester_name,
                                     fg='black', bg='white',
                                     width=self.col12_width)

        if self._initials_key_pressed_callback:
            self._initials_key_pressed_callback()  # pylint: disable=not-callable

    # -------------------
    ## translate tester initials into tester name
    #
    # @return None
    def _get_tester_name(self):
        if services.status.last_tester_initials in services.cfg.testers:
            self._tester_name = services.cfg.testers[services.status.last_tester_initials]
        else:
            self._tester_name = 'unknown'
        services.logger.user(f'tester name: "{self._tester_name}"')

    # -------------------
    ## create empty row
    #
    # @param row  the row to place it in
    # @return None
    def _create_empty_row(self, row):
        lbl = tk.Label(self._frame, text=' ',
                       width=0,
                       font=self.common_font,
                       fg='lightgrey',
                       bg='lightgrey',
                       highlightbackground='lightgrey', highlightthickness=1)
        lbl.configure(state='disabled')  # make it readonly
        lbl.grid(row=row, column=0, sticky='W')

    # -------------------
    ## crate a label widget
    #
    # @param text    the text for the widget
    # @param row     the row to place it in
    # @param col     the column to place it in
    # @param height  the height (in lines) of the lbl widget
    # @return None
    def _create_cell_lbl(self, text, row, col, height=1):
        lbl = tk.Label(self._frame,
                       text=text,
                       anchor='sw',
                       height=height,
                       width=self.col0_txt_width,
                       font=self.small_font,
                       fg='black',
                       bg='lightgrey')
        lbl.grid(row=row, column=col, sticky='EWS', pady=(5, 0), ipady=2)

    # -------------------
    ## btn that indicates the tester wants to abort the test
    #
    # @param row     the row to place it in
    # @param col     the column to place it in
    # @param fn_handle_click  the callback for handling a click
    # @return None
    def _create_abort_btn(self, row, col, fn_handle_click):
        btn = tk.Button(self._frame, text='Abort', command=fn_handle_click,
                        font=self.common_font,
                        fg='black',
                        bg='red',
                        width=self.col0_btn_width)
        btn.grid(row=row, column=col, sticky='W')

    # -------------------
    ## Create the Next btn. The tester presses
    # it to go to the next protocol step.
    #
    # @param row     the row to place it in
    # @param col     the column to place it in
    # @param fn_handle_click  the callback for handling a click
    # @return None
    def _create_next_btn(self, row, col, fn_handle_click):
        # initially disabled
        self._btn_next = tk.Button(self._frame, text='Next', command=fn_handle_click,
                                   state='disabled',
                                   font=self.common_font,
                                   fg='black',
                                   bg='darkblue',
                                   disabledforeground='lightgrey',
                                   width=self.col0_btn_width)
        self._btn_next.grid(row=row, column=col, sticky='W')

    # -------------------
    ## enable the Next btn
    #
    # @return None
    def _next_enable(self):
        self._btn_next['state'] = 'normal'

    # -------------------
    ## disable the Next btn
    #
    # @return None
    def _next_disable(self):
        self._btn_next['state'] = 'disabled'

    # -------------------
    ## create message box label. Filled with messages to the tester
    #
    # @param row  the row to place it in
    # @param msg  the message to display
    # @return None
    def _create_message_box(self, row, msg):
        self._lbl_message = tk.Label(self._frame,
                                     text=msg,
                                     anchor='center',
                                     wraplength=self.width,
                                     font=self.small_font,
                                     fg='black',
                                     bg='lightgrey',
                                     highlightbackground='black', highlightthickness=1)
        self._set_message('black', 'Press Can\'t do or perform the action and press Done')
        self._lbl_message.grid(row=row, column=0, sticky='NEW', columnspan=3, pady=(0, 5), ipady=10)

    # -------------------
    ## set the text of the message box
    #
    ## @param color  the color to print the text in
    ## @param text   the text to display
    # @return None
    def _set_message(self, color, text):
        self._lbl_message.config(text=text, fg=color)



# -------------------
## Holds information used to format a value
class ResultFormatter(dict):
    # -------------------
    ## constructor
    def __init__(self):
        super().__init__()

        ## the tag to use for the actual value
        self.actual_tag = None

        ## flag indicating to format numeric value as hex
        self.as_hex = False

        ## user provided format to use
        self.format = None

    # TODO use json content in cfg.json to create a default formatter

    # -------------------
    ## required for JSON to handle this class as a dictionary
    #
    # @param key      the name of the attribute to set
    # @param value    the value of the attribute
    # @return None
    def __setattr__(self, key, value):
        self[key] = value

    # -------------------
    ## required for JSON to handle this class as a dictionary
    #
    # @param key      the name of the attribute to get
    # @return the value of the named attribute
    def __getattr__(self, key):
        return self[key]



import os

from .gen_trace_matrix_base import GenTraceMatrixBase
from ..common.gen_base_txt import GenBaseTxt
from medver_pytest import services
from medver_pytest.constants import Constants


# -------------------
## Generates a Trace Matrix report in text format
class GenTraceMatrixTxt(GenBaseTxt, GenTraceMatrixBase):
    # -------------------
    ## constructor
    #
    # @param matrix   the data to use
    def __init__(self, matrix):
        GenBaseTxt.__init__(self)
        GenTraceMatrixBase.__init__(self, matrix)

        ## holds path to the output file
        self._path = os.path.join(services.cfg.outdir, f'{Constants.trace_fname}.txt')

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()

        with open(self._path, 'w', encoding='UTF-8') as self._fp:
            self._gen_test_run_details()
            self._gen_title('Trace Matrix')

            self._load_matrix_requirements()
            ## see base class doc
            self._report_requirements(self._report_header, self._report_desc, self._report_details)

    # -------------------
    ## callback to report the current requirement and tester info
    #
    # @param req_id   the requirement id to report
    # @param info     the info to report
    # @return None
    def _report_header(self, req_id, info):
        self._fp.write(f'req id: {req_id} {info}\n')

    # -------------------
    ## callback to report the current requirement's description (if available)
    #
    # @param desc     the description to report
    # @return None
    def _report_desc(self, desc):
        self._fp.write(f'   desc : {desc}\n')

    # -------------------
    ## callback to report protocol id and info for the current requirement
    #
    # @param proto_id     the protocol id to report
    # @param proto_info   the protocol info to report
    # @return None
    def _report_details(self, proto_id, proto_info):
        if proto_id is None and proto_info is None:
            self._fp.write('\n')
        elif proto_id is None:
            self._fp.write(f'   {proto_info}\n')
        else:
            self._fp.write(f'   proto: {proto_id} {proto_info}\n')



from .protocol.gen_tp_docx import GenTpDocx
from .protocol.gen_tp_minhtml import GenTpMinHtml
from .protocol.gen_tp_pdf import GenTpPdf
from .protocol.gen_tp_txt import GenTpTxt
from .summary.gen_summary_docx import GenSummaryDocx
from .summary.gen_summary_minhtml import GenSummaryMinHtml
from .summary.gen_summary_pdf import GenSummaryPdf
from .summary.gen_summary_txt import GenSummaryTxt
from .trace_matrix.gen_trace_matrix_docx import GenTraceMatrixDocx
from .trace_matrix.gen_trace_matrix_minhtml import GenTraceMatrixMinHtml
from .trace_matrix.gen_trace_matrix_pdf import GenTraceMatrixPdf
from .trace_matrix.gen_trace_matrix_txt import GenTraceMatrixTxt
from medver_pytest import services


# -------------------
## Generates all reports
# currently generates 3 types of reports:
#   * Protocol with results
#   * Trace matrix report
#   * Summary report
class Report:
    # -------------------
    ## constructor
    def __init__(self):
        pass

    # -------------------
    ## run all reports
    #
    # @return None
    def report(self):
        self._report_protocols()
        self._report_trace_matrix()
        self._report_summary()

    # -------------------
    ## run all protocol and reports
    #
    # @return None
    def _report_protocols(self):
        protocols = services.storage.get_protocols()

        # test protocol (no results): txt, pdf and docx
        services.cfg.page_info.set_tp_protocol_cfg()
        if 'txt' in services.cfg.report_types:
            rep = GenTpTxt(protocols, do_results=False)
            rep.gen()

        if 'pdf' in services.cfg.report_types:
            rep = GenTpPdf(protocols, do_results=False)
            rep.gen()

        if 'docx' in services.cfg.report_types:
            rep = GenTpDocx(protocols, do_results=False)
            rep.gen()

        if 'min-html' in services.cfg.report_types:
            rep = GenTpMinHtml(protocols, do_results=False)
            rep.gen()

        # reports with results: txt, pdf and docx
        services.cfg.page_info.set_tp_report_cfg()
        if 'txt' in services.cfg.report_types:
            rep = GenTpTxt(protocols, do_results=True)
            rep.gen()

        if 'pdf' in services.cfg.report_types:
            rep = GenTpPdf(protocols, do_results=True)
            rep.gen()

        if 'docx' in services.cfg.report_types:
            rep = GenTpDocx(protocols, do_results=True)
            rep.gen()

        if 'min-html' in services.cfg.report_types:
            rep = GenTpMinHtml(protocols, do_results=True)
            rep.gen()

    # -------------------
    ## run all trace matrix reports
    #
    # @return None
    def _report_trace_matrix(self):
        matrix = services.storage.get_trace()
        services.cfg.page_info.set_trace_cfg()

        if 'txt' in services.cfg.report_types:
            rep = GenTraceMatrixTxt(matrix)
            rep.gen()

        if 'pdf' in services.cfg.report_types:
            rep = GenTraceMatrixPdf(matrix)
            rep.gen()

        if 'docx' in services.cfg.report_types:
            rep = GenTraceMatrixDocx(matrix)
            rep.gen()

        if 'min-html' in services.cfg.report_types:
            rep = GenTraceMatrixMinHtml(matrix)
            rep.gen()

    # -------------------
    ## run all summary reports
    #
    # @return None
    def _report_summary(self):
        summary = services.storage.get_summary()
        services.cfg.page_info.set_summary_cfg()
        if 'txt' in services.cfg.report_types:
            rep = GenSummaryTxt(summary)
            rep.gen()

        if 'pdf' in services.cfg.report_types:
            rep = GenSummaryPdf(summary)
            rep.gen()

        if 'docx' in services.cfg.report_types:
            rep = GenSummaryDocx(summary)
            rep.gen()

        if 'min-html' in services.cfg.report_types:
            rep = GenSummaryMinHtml(summary)
            rep.gen()



import glob
import json
import os
from collections import OrderedDict

from .storage import Storage
from medver_pytest import services


# -------------------
## Holds base functions for use with file storage
class StorageFileBase(Storage):
    # -------------------
    ## constructor
    def __init__(self):
        ## holds the path to the protocol json file
        self._protocol_path = None
        ## holds the path to the trace matrix json file
        self._trace_path = None
        ## holds the path to the summary json file
        self._summary_path = None

    # -------------------
    ## (overridden) initialize
    #
    # @return None
    def init(self):
        # coverage: overridden methods not called
        pass  # pragma: no cover

    # -------------------
    ## (overridden) initialize
    #
    # @return None
    def term(self):
        # coverage: overridden methods not called
        pass  # pragma: no cover

    # -------------------
    ## save protocol information to local file
    #
    # @param protocols    the info to save
    # @return None
    def save_protocol(self, protocols):
        if services.cfg.report_mode or services.cfg.test_script is None:
            return

        self._protocol_path = os.path.join(services.cfg.outdir, f'{services.cfg.test_script}_protocol.json')
        with open(self._protocol_path, 'w', encoding='utf=8') as fp:
            json.dump(protocols, fp, ensure_ascii=True, indent=2)

    # -------------------
    ## get protocol information from local file
    #
    # @return the protocol info
    def get_protocols(self):
        protocols = {}
        for f in glob.glob(os.path.join(services.cfg.outdir, '*_protocol.json')):
            with open(f, 'r', encoding='utf=8') as fp:
                services.logger.ok(f'loading protocol file: {f}')
                curr_protos = json.load(fp)

                for proto_id, protocol in curr_protos.items():
                    if proto_id in protocols:
                        curr_loc = protocol['location']
                        prev_loc = protocols[proto_id]['location']
                        services.logger.err(f'{proto_id} protocol id is already in use:')
                        services.logger.err(f'   location: {curr_loc}')
                        services.logger.err(f'   previous: {prev_loc}')
                    else:
                        protocols[proto_id] = protocol
        return OrderedDict(sorted(protocols.items()))

    # -------------------
    ## save trace matrix information to local file
    #
    # @param trace    the info to save
    # @return None
    def save_trace(self, trace):
        if services.cfg.report_mode or services.cfg.test_script is None:
            return

        self._trace_path = os.path.join(services.cfg.outdir, f'{services.cfg.test_script}_trace.json')
        with open(self._trace_path, 'w', encoding='utf=8') as fp:
            json.dump(trace, fp, ensure_ascii=True, indent=2)

    # -------------------
    ## save trace matrix information get from the local file
    #
    # @return the trace matrix info
    def get_trace(self):
        jlist = {}
        for f in glob.glob(os.path.join(services.cfg.outdir, '*_trace.json')):
            with open(f, 'r', encoding='utf=8') as fp:
                items = json.load(fp)

                # TODO callback into trace object
                for key in items.keys():
                    if key not in jlist:
                        jlist[key] = []

                    for item in items[key]:
                        jlist[key].append(item)
        return jlist

    # -------------------
    ## save summary information to local file
    #
    # @param summary    the info to save
    # @return None
    def save_summary(self, summary):
        if services.cfg.report_mode or services.cfg.test_script is None:
            return

        self._summary_path = os.path.join(services.cfg.outdir, f'{services.cfg.test_script}_summary.json')
        with open(self._summary_path, 'w', encoding='utf=8') as fp:
            json.dump(summary, fp, ensure_ascii=True, indent=2)

    # -------------------
    ## get summary information from the local file
    #
    # @return the summary information retrieved
    def get_summary(self):
        jlist = {
            'reqids': {},
            'protoids': {},
        }
        for f in glob.glob(os.path.join(services.cfg.outdir, '*_summary.json')):
            with open(f, 'r', encoding='utf=8') as fp:
                items = json.load(fp)

                # TODO callback into summary object
                reqids = items['reqids']
                for key in reqids.keys():
                    if key in jlist['reqids']:
                        jlist['reqids'][key]['count'] += 1
                        if reqids[key]['result'] == 'FAIL':
                            jlist['reqids'][key]['result'] = 'FAIL'
                    else:
                        jlist['reqids'][key] = reqids[key]

                protoids = items['protoids']
                for key in protoids.keys():
                    if key in jlist['protoids']:
                        services.logger.err(f'{key} protocol id is already in use, skipping')
                    else:
                        jlist['protoids'][key] = protoids[key]

        return jlist



import os

from reportlab.lib import colors
from reportlab.lib.units import inch
from reportlab.platypus import Paragraph
from reportlab.platypus import Table
from reportlab.platypus import TableStyle

from .gen_trace_matrix_base import GenTraceMatrixBase
from ..common.gen_base_pdf import GenBasePdf
from medver_pytest import services
from medver_pytest.constants import Constants


# -------------------
## Generates a Trace Matrix report in PDF format
class GenTraceMatrixPdf(GenBasePdf, GenTraceMatrixBase):
    # -------------------
    ## constructor
    #
    # @param matrix   the data to use
    def __init__(self, matrix):
        GenBasePdf.__init__(self)
        GenTraceMatrixBase.__init__(self, matrix)

        ## holds reference to the current table being generated
        self._tbl = None
        ## holds current reqmt id for the row
        self._row_id = None
        ## holds current reqmt info for the row
        self._row_info = None

        path = os.path.join(services.cfg.outdir, f'{Constants.trace_fname}.pdf')
        self._doc_init(path)

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()

        self._gen_test_run_details()
        self._gen_title('Trace Matrix')

        self._gen_trace_table()

        self._build()

    # -------------------
    ## generate a trace matrix table
    #
    # @return None
    def _gen_trace_table(self):
        self._tbl = []
        self._tbl.append(
            [
                Paragraph('<b>Req. Id</b>'),
                Paragraph('<b>Protocol</b>'),
            ]
        )

        self._load_matrix_requirements()
        ## see base class doc
        self._report_requirements(self._report_header, self._report_desc, self._report_details)

        self._gen_table()

    # -------------------
    ## callback to report the current requirement and tester info
    #
    # @param req_id   the requirement id to report
    # @param info     the info to report
    # @return None
    def _report_header(self, req_id, info):
        if self._row_id is not None:
            self._tbl.append([self._row_id, self._row_info])

        self._row_id = Paragraph(f'{req_id}\n{info}', self._style_sheet['CenterPara'])
        self._row_info = []

    # -------------------
    ## callback to report the current requirement's description (if available)
    #
    # @param desc     the description to report
    # @return None
    def _report_desc(self, desc):
        self._row_info.append(Paragraph(f'<i><b>Desc:</b> {desc}</i>',
                                        self._style_sheet['ExtraSpace']))

    # -------------------
    ## callback to report protocol id and info for the current requirement
    #
    # @param proto_id     the protocol id to report
    # @param proto_info   the protocol info to report
    # @return None
    def _report_details(self, proto_id, proto_info):
        if proto_id is None and proto_info is None:
            self._tbl.append([self._row_id, self._row_info])
            self._row_id = None
            self._row_info = None
        elif proto_id is None:
            self._row_info.append(Paragraph(f'{proto_info}',
                                            self._style_sheet['LeftPara']))
        else:
            self._row_info.append(Paragraph(f'{proto_id} {proto_info}',
                                            self._style_sheet['LeftPara']))

    # -------------------
    ## generate the trace matrix table
    #
    # @return None
    def _gen_table(self):
        tbl_widths = [
            0.75 * inch,  # reqid
            6.25 * inch,  # proto_info
        ]

        t = Table(self._tbl,
                  colWidths=tbl_widths,
                  spaceBefore=0.25 * inch,
                  spaceAfter=0.25 * inch,
                  )

        t.setStyle(TableStyle(
            [
                # borders
                ('BOX', (0, 0), (-1, -1), 0.25, colors.black),
                ('INNERGRID', (0, 0), (-1, -1), 0.25, colors.black),

                # top row is a title row
                ('BACKGROUND', (0, 0), (1, 0), colors.lightblue),

                # "reqid" column
                ('VALIGN', (0, 0), (0, -1), 'MIDDLE'),
                ('ALIGN', (0, 0), (0, -1), 'CENTER'),

                # "protocol" column
                ('VALIGN', (1, 0), (1, -1), 'MIDDLE'),
                ('ALIGN', (1, 0), (1, -1), 'LEFT'),
            ]
        ))
        self._elements.append(t)



from abc import ABC
from abc import abstractmethod

from medver_pytest import services


# -------------------
## Abstract base class for storage functions
class Storage(ABC):
    # -------------------
    ## factory to generate the child class given the storage type
    #
    # @return a storage object
    @classmethod
    def factory(cls):
        if services.cfg.storage_type == 'local':
            from .storage_dev import StorageDev

            return StorageDev()
        elif services.cfg.storage_type == 'shared':
            from .storage_shared_drive import StorageSharedDrive

            return StorageSharedDrive()

        services.abort(f'unknown storage type: {services.cfg.storage_type}')

    # -------------------
    ## (abstract) initialize
    #
    # @return None
    @abstractmethod
    def init(self):
        # coverage: abstract methods not called
        pass  # pragma: no cover

    # -------------------
    ## (abstract) terminate
    #
    # @return None
    @abstractmethod
    def term(self):
        # coverage: abstract methods not called
        pass  # pragma: no cover

    # -------------------
    ## (abstract) save the protocol information
    #
    # @param protocols  the protocol information to save
    # @return None
    @abstractmethod
    def save_protocol(self, protocols):
        # coverage: abstract methods not called
        pass  # pragma: no cover

    # -------------------
    ## (abstract) get saved protocol information
    #
    # @return None
    @abstractmethod
    def get_protocols(self):
        # coverage: abstract methods not called
        pass  # pragma: no cover

    # -------------------
    ## (abstract) save the trace matrix information
    #
    # @param trace   the trace matrix information to save
    # @return None
    @abstractmethod
    def save_trace(self, trace):
        # coverage: abstract methods not called
        pass  # pragma: no cover

    # -------------------
    ## (abstract) get saved trace matrix information
    #
    # @return None
    @abstractmethod
    def get_trace(self):
        # coverage: abstract methods not called
        pass  # pragma: no cover

    # -------------------
    ## (abstract) save the summary information
    #
    # @param summary   the summary information to save
    # @return None
    @abstractmethod
    def save_summary(self, summary):
        # coverage: abstract methods not called
        pass  # pragma: no cover

    # -------------------
    ## (abstract) get saved summary information
    #
    # @return None
    @abstractmethod
    def get_summary(self):
        # coverage: abstract methods not called
        pass  # pragma: no cover



import json
import os
import re

import jsmin

from medver_pytest.stdout_logger import StdoutLogger as logger


# -------------------
## Holds content for IUV requirements
class IUVRequirements:
    # -------------------
    ## constructor
    def __init__(self):
        ## path to the IUV requirements json file
        self._reqmts_path = os.path.join('iuv', 'iuv_srs.json')
        ## the IUV requirements
        self._reqmts = None

        ## path to the IUV trace json file
        self._trace_path = os.path.join('out', 'iuv', 'iuv_reqids.json')
        ## the IUV trace info
        self._trace = {}

    # -------------------
    ## initialize
    #
    # @return None
    def init(self):
        with open(self._reqmts_path, 'r', encoding='utf-8') as fp:
            cleanj = jsmin.jsmin(fp.read())
            self._reqmts = json.loads(cleanj)

        self.load_trace()

    # -------------------
    ## check the given reqid is valid
    #
    # @param reqid       the reqid to check
    # @param location    the location of the reqid
    # @return None if the reqid is ok, otherwise a message to print
    def check(self, reqid, location):
        msg = None
        # check reqid is named
        if reqid is None or reqid == '':
            msg = f'IUV: missing reqid: {location}'

        # check reqid is valid format
        if not re.search(r'^IUV-(\d+)', reqid):
            msg = f'IUV: bad reqid format: "{reqid}"'

        return msg

    # -------------------
    ## save iuv_verify() info to the trace json
    #
    # @param reqids    the reqid list for this call
    # @param result    whether the call passed or faild
    # @param location  the iuv_verify() location
    # @return None
    def save_reqids_to_trace(self, reqids, result, location):
        for reqid in reqids:
            if reqid not in self._trace:
                self._trace[reqid] = {
                    'passed': True,
                    'locations': [],
                }
            self._trace[reqid]['passed'] = result and self._trace[reqid]['passed']
            self._trace[reqid]['locations'].append(location)

        self.save_trace()

    # -------------------
    ## get desc for given requirement id
    #
    # @param reqid   the requirment id
    # @return description of the reqid
    def get(self, reqid):
        return self._reqmts[reqid]['desc']

    # -------------------
    ## load the trace file
    #
    # @return None
    def load_trace(self):
        if os.path.isfile(self._trace_path):
            with open(self._trace_path, 'r', encoding='utf-8') as fp:
                self._trace = json.load(fp)
        else:
            self._trace = {}

    # -------------------
    ## save trace content
    #
    # @return None
    def save_trace(self):
        if self._trace != {}:
            with open(self._trace_path, 'w', encoding='utf-8') as fp:
                json.dump(self._trace, fp, indent=4)

    # -------------------
    ## report summary of tested requirements and which passed/failed
    #
    # @param log_it  a callback function to write to the IUV log
    # @return None
    def report_trace(self, log_it):
        logger.line('IUV Report:')
        self.load_trace()

        results = self._trace

        total_tested_reqmts = 0
        total_tested_passed = 0
        total_passed_locations = 0
        total_tested_failed = 0
        total_failed_locations = 0
        results_keys = results.keys()
        for reqid in sorted(results_keys):
            info = results[reqid]
            total_tested_reqmts += 1
            if info['passed']:
                total_tested_passed += 1
                total_passed_locations += len(info['locations'])
            else:
                total_tested_failed += 1
                total_failed_locations += len(info['locations'])

        log_it('Total tested requirements', total_tested_reqmts, total_tested_reqmts)
        log_it('Total passed reqmts', total_tested_passed, total_tested_reqmts)
        log_it('Total passed locations', total_passed_locations)
        log_it('Total failed reqmts', total_tested_failed, total_tested_reqmts)
        log_it('Total failed locations', total_failed_locations)
        log_it('')

    # -------------------
    ## report list of IUV requirements that were not tested
    #
    # @param log_it  a callback function to write to the IUV log
    # @return None
    def report_missing(self, log_it):
        log_it('IUV Requirements not tested:')
        total_reqmts = 0
        total_tested = 0
        total_not_tested = 0
        for reqid in sorted(self._reqmts.keys()):
            total_reqmts += 1
            if reqid in self._trace.keys():
                total_tested += 1
            else:
                log_it(f'   {reqid: <8} {self.get(reqid)}')
                total_not_tested += 1

        if total_not_tested == 0:
            log_it('   None')

        log_it('')

        log_it('IUV Requirements summary:')
        log_it('Total requirements', total_reqmts, total_reqmts)
        log_it('Total tested', total_tested, total_reqmts)
        log_it('Total not tested', total_not_tested, total_reqmts)
        log_it('')



import docx
import docx.enum
import docx.enum.section
import docx.enum.style
import docx.enum.table
import docx.enum.text
import docx.shared
from docx.oxml.ns import qn
from docx.oxml.shared import OxmlElement

from medver_pytest import services
from medver_pytest.constants import Constants
from medver_pytest.utils import Utils


# -------------------
## Base class for generating a msword docx
class GenBaseDocx:  # pylint: disable=too-few-public-methods

    # -------------------
    ## constructor
    def __init__(self):
        ## holds reference to the current PDF doc
        self._doc = None

    # -------------------
    ## initialize document
    #
    # @return None
    def _doc_init(self):
        self._set_page_layout()
        self._add_styles()
        self._gen_headers_footers()

    # -------------------
    ## set page layout to landscape/portrait
    #
    # @return None
    def _set_page_layout(self):
        section = self._doc.sections[0]
        self._set_section_orientation(section)

        section.header_distance = docx.shared.Inches(0.0)
        section.left_margin = docx.shared.Inches(0.5)
        section.right_margin = docx.shared.Inches(0.5)
        section.top_margin = docx.shared.Inches(0.25)
        section.bottom_margin = docx.shared.Inches(0.25)

    # -------------------
    ## add styles needed for various tables
    #
    # @return None
    def _add_styles(self):
        # Note: use keep_with_next to minimize splitting across pages
        self._add_style_table_header()
        self._add_style_table_cell()
        self._add_style_table_desc_cell()
        self._add_style_monospace()
        self._add_style_list_item()

    # -------------------
    ## add a paragraph style to the document, set common values
    #
    # @param name  the style name
    # @return the style
    def _add_style(self, name):
        style = self._doc.styles.add_style(name, docx.enum.style.WD_STYLE_TYPE.PARAGRAPH)  # pylint: disable=no-member
        style.paragraph_format.line_spacing_rule = docx.enum.text.WD_LINE_SPACING.SINGLE  # pylint: disable=no-member
        style.paragraph_format.alignment = self._align_left()
        style.paragraph_format.keep_with_next = True
        return style

    # -------------------
    ## add a paragraph style for table headings (bold)
    #
    # @return None
    def _add_style_table_header(self):
        style = self._add_style('ver_table1_header')
        style.base_style = self._doc.styles['Normal']
        style.font.name = 'Arial'
        style.font.size = docx.shared.Pt(8)
        style.font.bold = True
        style.paragraph_format.space_before = docx.shared.Inches(0.08)
        style.paragraph_format.space_after = docx.shared.Inches(0.08)

    # -------------------
    ## add a paragraph style for table cells
    #
    # @return None
    def _add_style_table_cell(self):
        style = self._add_style('ver_table1_cell')
        style.base_style = self._doc.styles['Normal']
        style.font.name = 'Arial'
        style.font.size = docx.shared.Pt(8)
        style.font.bold = False
        style.paragraph_format.space_before = docx.shared.Inches(0.0)
        style.paragraph_format.space_after = docx.shared.Inches(0.0)

    # -------------------
    ## add a paragraph style for descriptions in tables
    #
    # @return None
    def _add_style_table_desc_cell(self):
        style = self._add_style('ver_table1_desc_cell')
        style.base_style = self._doc.styles['Normal']
        style.font.name = 'Arial'
        style.font.size = docx.shared.Pt(8)
        style.font.bold = False
        style.paragraph_format.space_before = docx.shared.Inches(0.08)
        style.paragraph_format.space_after = docx.shared.Inches(0.0)

    # -------------------
    ## add a paragraph style with monospace font
    #
    # @return None
    def _add_style_monospace(self):
        style = self._add_style('ver_monospace')
        style.base_style = self._doc.styles['Normal']
        style.font.name = 'Courier New'
        style.font.size = docx.shared.Pt(7)
        style.font.bold = True
        style.paragraph_format.space_before = docx.shared.Inches(0.0)
        style.paragraph_format.space_after = docx.shared.Inches(0.0)

    # -------------------
    ## add a paragraph style for lists
    #
    # @return None
    def _add_style_list_item(self):
        style = self._add_style('ver_list_item')
        style.base_style = self._doc.styles['List Bullet']
        style.font.name = 'Arial'
        style.font.size = docx.shared.Pt(8)
        style.font.bold = False
        style.paragraph_format.space_before = docx.shared.Inches(0.08)
        style.paragraph_format.space_after = docx.shared.Inches(0.0)

    # -------------------
    ## add before/after space to a table
    #
    # @param size   the font size to use
    # @return None
    def _tbl_add_space(self, size):
        para = self._doc.add_paragraph()
        run = para.add_run(' ')  # must have a space!
        run.font.size = docx.shared.Pt(size)

        # 20ths of a point
        # tbl._tblPr.set(qn('w:bottomFromText'), '1440')  # pylint: disable=protected-access

        # logger.dbg(f'"{tbl._tblPr}"')  # pylint: disable=protected-access
        # tbl._tblPr.append(gap)

        # tblPr = tbl._tbl.tblPr
        # floating = OxmlElement('w:tblpPr')  # floating table properties
        # floating.set(qn('w:topFromText'), '720')  # <--------- relevant values
        # floating.set(qn('w:bottomFromText'), '720')  # <--------- relevant values
        # floating.set(qn('w:tblpY'), '721')  # <--------- relevant values
        # floating.set(qn('w:vertAnchor'), 'text')
        # tblPr.append(floating)  # floating table properties
        #
        # overlap = OxmlElement('w:tblOverlap')  # overlap properties
        # overlap.set(qn('w:val'), 'never')
        # tblPr.append(overlap)

    # -------------------
    ## set the background of the given cell to light blue
    #
    # @param cell  the cell to change
    # @return None
    def _set_cell_blue_background(self, cell):
        shd = OxmlElement('w:shd')
        shd.set(qn('w:val'), 'clear')
        shd.set(qn('w:color'), 'auto')
        shd.set(qn('w:fill'), 'ADD8E6')  # light blue
        cell._tc.get_or_add_tcPr().append(shd)  # pylint: disable=protected-access

    # -------------------
    ## set the background of the given cell to light grey
    #
    # @param cell  the cell to change
    # @return None
    def _set_cell_grey_background(self, cell):
        shd = OxmlElement('w:shd')
        shd.set(qn('w:val'), 'clear')
        shd.set(qn('w:color'), 'auto')
        shd.set(qn('w:fill'), 'D9D9D9')  # light grey
        cell._tc.get_or_add_tcPr().append(shd)  # pylint: disable=protected-access

    # -------------------
    ## generate test run information
    #
    # @return None
    def _gen_test_run_details(self):
        self._doc.add_heading('Test Run Details', level=3)

        line = f"{'Test Run Type'}: {services.cfg.test_run_type}"
        self._doc.add_paragraph(line, style='List Bullet')

        line = f"{'Test Run ID'}: {services.cfg.test_run_id}"
        self._doc.add_paragraph(line, style='List Bullet')

        dts = Utils.get_dts(use_cfg_fmt=True)
        line = f"{'Document Generated'}: {dts}"
        self._doc.add_paragraph(line, style='List Bullet')

        line = f"{'medver-pytest version'}: v{Constants.version}"
        self._doc.add_paragraph(line, style='List Bullet')

    # -------------------
    ## generate title
    #
    # @param title  the title to draw
    # @return None
    def _gen_title(self, title):
        self._doc.add_heading(title, level=3)

    # -------------------
    ## generate headers and footers
    #
    # @return None
    def _gen_headers_footers(self):
        self._gen_headers()
        self._gen_footers()

    # -------------------
    ## generate headers
    #
    # @return None
    def _gen_headers(self):
        header = self._doc.sections[0].header
        if services.cfg.page_info.orientation == 'portrait':
            tbl = header.add_table(rows=1, cols=3, width=docx.shared.Inches(7.5))
        else:
            tbl = header.add_table(rows=1, cols=3, width=docx.shared.Inches(10.0))
        tbl.style = 'Normal Table'
        self._align_table_center(tbl)

        edge_width = docx.shared.Inches(1.75)
        if services.cfg.page_info.orientation == 'portrait':
            # should add to 7.5"
            tbl.columns[0].width = edge_width
            tbl.columns[1].width = docx.shared.Inches(4.00)
            tbl.columns[2].width = edge_width
        else:
            # should add to 10"
            tbl.columns[0].width = edge_width
            tbl.columns[1].width = docx.shared.Inches(6.5)
            tbl.columns[2].width = edge_width

        row = tbl.rows[0].cells
        self._gen_hf_data(row, 'header')

        para = header.add_paragraph()
        self._insert_hr(para)

    # -------------------
    ## generate footers
    #
    # @return None
    def _gen_footers(self):
        footer = self._doc.sections[0].footer
        self._insert_hr(footer.paragraphs[0])
        if services.cfg.page_info.orientation == 'portrait':
            tbl = footer.add_table(rows=1, cols=3, width=docx.shared.Inches(7.5))
        else:
            tbl = footer.add_table(rows=1, cols=3, width=docx.shared.Inches(10.0))
        tbl.style = 'Normal Table'
        self._align_table_center(tbl)
        tbl.keep_with_next = True

        edge_width = docx.shared.Inches(1.75)
        if services.cfg.page_info.orientation == 'portrait':
            # should add to 7.5"
            tbl.columns[0].width = edge_width
            tbl.columns[1].width = docx.shared.Inches(4.00)
            tbl.columns[2].width = edge_width
        else:
            # should add to 10"
            tbl.columns[0].width = edge_width
            tbl.columns[1].width = docx.shared.Inches(6.5)
            tbl.columns[2].width = edge_width

        row = tbl.rows[0].cells
        self._gen_hf_data(row, 'footer')

    # -------------------
    ## generate headers/footer data
    #
    # @param row  the row for the header/footer data
    # @param tag  indicates header or footer
    # @return None
    def _gen_hf_data(self, row, tag):
        col = 0
        if services.cfg.page_info[tag].left == '<pageno>':
            self._add_page_number(row, col, self._align_left())
        else:
            txt = services.cfg.page_info[tag].left
            txt = txt.replace('<br/>', '\n')
            boldness = False
            if txt.find('<b>') != -1 or txt.find('</b>') != -1 or txt.find('<b/>') != -1:
                boldness = True
                txt = txt.replace('<b>', '')
                txt = txt.replace('</b>', '')
                txt = txt.replace('<b/>', '')
            row[col].paragraphs[0].add_run(txt).bold = boldness
            row[col].paragraphs[0].alignment = self._align_left()

        col = 1
        if services.cfg.page_info[tag].middle == '<pageno>':
            self._add_page_number(row, col, self._align_center())
        else:
            txt = services.cfg.page_info[tag].middle
            txt = txt.replace('<br/>', '\n')
            boldness = False
            if txt.find('<b>') != -1 or txt.find('</b>') != -1 or txt.find('<b/>') != -1:
                boldness = True
                txt = txt.replace('<b>', '')
                txt = txt.replace('</b>', '')
                txt = txt.replace('<b/>', '')
            row[col].paragraphs[0].add_run(txt).bold = boldness
            row[col].paragraphs[0].alignment = self._align_center()

        col = 2
        if services.cfg.page_info[tag].right == '<pageno>':
            self._add_page_number(row, col, self._align_left())
        else:
            txt = services.cfg.page_info[tag].right
            txt = txt.replace('<br/>', '\n')
            boldness = False
            if txt.find('<b>') != -1 or txt.find('</b>') != -1 or txt.find('<b/>') != -1:
                boldness = True
                txt = txt.replace('<b>', '')
                txt = txt.replace('</b>', '')
                txt = txt.replace('<b/>', '')
            row[col].paragraphs[0].add_run(txt).bold = boldness
            row[col].paragraphs[0].alignment = self._align_left()

    # -------------------
    ## add page number to header/footer
    #
    # @param row        the cell to add the page to
    # @param col        the col number to add the page to
    # @param alignment  the paragraph alignment to use
    # @return None
    def _add_page_number(self, row, col, alignment):
        run = row[col].paragraphs[0].add_run('Page ')

        elem = OxmlElement('w:fldChar')
        elem.set(qn('w:fldCharType'), 'begin')
        run._r.append(elem)  # pylint: disable=protected-access

        elem = OxmlElement('w:instrText')
        elem.set(qn('xml:space'), 'preserve')
        elem.text = "PAGE"
        run._r.append(elem)  # pylint: disable=protected-access

        elem = OxmlElement('w:fldChar')
        elem.set(qn('w:fldCharType'), 'end')
        run._r.append(elem)  # pylint: disable=protected-access

        row[col].paragraphs[0].alignment = alignment

    # -------------------
    ## insert a header paragraph
    #
    # @param paragraph the paragraph to add to the header
    # @return None
    def _insert_hr(self, paragraph):
        para = paragraph._p  # pylint: disable=protected-access
        ppr = para.get_or_add_pPr()
        pbdr = OxmlElement('w:pBdr')
        ppr.insert_element_before(pbdr,
                                  'w:shd', 'w:tabs', 'w:suppressAutoHyphens', 'w:kinsoku', 'w:wordWrap',
                                  'w:overflowPunct', 'w:topLinePunct', 'w:autoSpaceDE', 'w:autoSpaceDN',
                                  'w:bidi', 'w:adjustRightInd', 'w:snapToGrid', 'w:spacing', 'w:ind',
                                  'w:contextualSpacing', 'w:mirrorIndents', 'w:suppressOverlap', 'w:jc',
                                  'w:textDirection', 'w:textAlignment', 'w:textboxTightWrap',
                                  'w:outlineLvl', 'w:divId', 'w:cnfStyle', 'w:rPr', 'w:sectPr',
                                  'w:pPrChange'
                                  )
        bottom = OxmlElement('w:bottom')
        bottom.set(qn('w:val'), 'single')
        bottom.set(qn('w:sz'), '0')
        bottom.set(qn('w:space'), '0')
        bottom.set(qn('w:color'), 'auto')
        pbdr.append(bottom)

    # -------------------
    ## set alignment to left
    #
    # @return None
    def _align_left(self):
        return docx.enum.text.WD_PARAGRAPH_ALIGNMENT.LEFT  # pylint: disable=no-member

    # -------------------
    ## set alignment to center
    #
    # @return None
    def _align_center(self):
        return docx.enum.text.WD_PARAGRAPH_ALIGNMENT.CENTER  # pylint: disable=no-member

    # -------------------
    ## set table alignment to center
    #
    # @param tbl  the table to align
    # @return None
    def _align_table_center(self, tbl):
        tbl.alignment = docx.enum.table.WD_TABLE_ALIGNMENT.CENTER  # pylint: disable=no-member

    # -------------------
    ## set cell's vertical alignment to center
    #
    # @param cell the cell to align
    # @return None
    def _align_vertical_center(self, cell):
        cell.vertical_alignment = docx.enum.table.WD_CELL_VERTICAL_ALIGNMENT.CENTER  # pylint: disable=no-member

    # -------------------
    ## set cell's vertical alignment to top
    #
    # @param cell the cell to align
    # @return None
    def _align_vertical_top(self, cell):
        cell.vertical_alignment = docx.enum.table.WD_CELL_VERTICAL_ALIGNMENT.TOP  # pylint: disable=no-member

    # -------------------
    ## set the portrait/landscape orientation for the given section
    #
    # @param section the section to set
    # @return None
    def _set_section_orientation(self, section):
        if services.cfg.page_info.page_size == 'letter':
            # the default
            pass
        elif services.cfg.page_info.page_size == 'a4':
            section.page_height = docx.shared.Mm(297)
            section.page_width = docx.shared.Mm(210)
            section.left_margin = docx.shared.Mm(25.4)
            section.right_margin = docx.shared.Mm(25.4)
            section.top_margin = docx.shared.Mm(25.4)
            section.bottom_margin = docx.shared.Mm(25.4)
            section.header_distance = docx.shared.Mm(12.7)
            section.footer_distance = docx.shared.Mm(12.7)
        else:
            services.abort(f'unknown page_size: {services.cfg.page_info.page_size}')

        if services.cfg.page_info.orientation == 'portrait':
            section.orientation = docx.enum.section.WD_ORIENTATION.PORTRAIT  # pylint: disable=no-member
            # no rotate needed
        else:
            section.orientation = docx.enum.section.WD_ORIENTATION.LANDSCAPE  # pylint: disable=no-member
            # rotate page
            new_width = section.page_width
            new_height = section.page_height
            section.page_width = new_height
            section.page_height = new_width



import os

from .gen_summary_base import GenSummaryBase
from ..common.gen_base_minhtml import GenBaseMinHtml
from medver_pytest import services
from medver_pytest.constants import Constants


# -------------------
## Generates a Summary report in min-html format
class GenSummaryMinHtml(GenSummaryBase, GenBaseMinHtml):
    # -------------------
    ## constructor
    #
    # @param summary    the summary data to use
    def __init__(self, summary):
        GenBaseMinHtml.__init__(self)
        GenSummaryBase.__init__(self, summary)

        ## holds path to the output file
        self._path = os.path.join(services.cfg.outdir, f'{Constants.summary_fname}-min.html')

        ## holds file pointer to the output file
        self._fp = None

    # -------------------
    ## generate the report
    #
    # @return None
    def gen(self):
        self._gen_init()

        with open(self._path, 'w', encoding='UTF-8') as self._fp:
            self._fp.write('<header id="ver-summary-title">\n')
            self._fp.write('    <h2 class="title">Automated Test Summary</h2>\n')
            self._fp.write('</header>\n')
            self._fp.write('\n')

            self._gen_test_run_details()
            self._fp.write('\n')

            self._gen_title('Summary', arg_id='Summary')
            self._gen_req_summary()
            self._gen_blank_line()
            self._gen_proto_summary()

    # -------------------
    ## generate requirements summary
    #
    # @return None
    def _gen_req_summary(self):
        self._gen_section_open('Requirements Summary')

        # TODO @@@ NEW table?
        # total requirements
        if self._requirements.is_found:
            total = self._requirements.total
            self._write_row('#Requirements', total, total)

            count = self._requirements.total_manual
            self._write_row('Manual Requirements', count, total)

            count = self._requirements.total_both
            self._write_row('Manual & Auto Requirements', count, total)

            total = self._requirements.total_automated
        else:
            total = len(self._summary['reqids'])

        self._write_row('Automated Requirements', total, total)

        # count failing/passing reqmts
        num_failing, num_passing, num_invalid, num_missing = self._count_requirements(total, self._report_invalid)
        self._write_row('Invalid', num_invalid, total, indent=True)
        self._write_row('PASS', num_passing, total, indent=True)
        self._write_row('FAIL', num_failing, total, indent=True)
        self._write_row('Not tested', num_missing, total, indent=True)

        self._gen_section_close()

    # -------------------
    ## callback to report any invalid requirements
    #
    # @param reqid  the reqid to report
    # @return None
    def _report_invalid(self, reqid):
        msg = f'   {reqid} not found in {services.cfg.reqmt_json_path}'
        services.logger.warn(msg)
        self._fp.write(f'{msg}\n')

    # -------------------
    ## generate protocol summary
    #
    # @return None
    def _gen_proto_summary(self):
        self._gen_section_open('Protocol Summary')

        # total protocols
        total = len(self._summary['protoids'])
        self._write_row('#Protocols', total, total)

        # count failing/passing reqmts
        num_failing, num_passing = self._count_protocols()
        self._write_row('PASS', num_passing, total, indent=True)
        self._write_row('FAIL', num_failing, total, indent=True)

        self._gen_section_close()

    # -------------------
    ## generates the start of a sections table
    #
    # @param title   the title of the table
    # @returns None
    def _gen_section_open(self, title):
        self._gen_table_open(css_class='ver-summary-table')
        # heading1
        self._gen_thead_open()
        self._gen_tr_open(css_class='ver-summary-hdg1')
        self._gen_th_strong(title)
        self._gen_th_strong()
        self._gen_th_strong()
        self._gen_tr_close()
        self._gen_thead_close()

        self._gen_tbody_open()
        # heading2
        self._gen_tr_open(css_class='ver-summary-hdg2')
        self._gen_td_strong(text=None, css_class='ver-summary-hdg-col1')
        self._gen_td_strong('Count', css_class='ver-summary-hdg-col2')
        self._gen_td_strong('Percentage', css_class='ver-summary-hdg-col3')
        self._gen_tr_close()

    # -------------------
    ## closes body and table of the current summary section
    #
    # @returns None
    def _gen_section_close(self):
        self._gen_tbody_close()
        self._gen_table_close()

    # -------------------
    ## write a row to the given file
    #
    # @param msg         the line header to write
    # @param count       the value to write
    # @param total       the total count
    # @param css_class   the css class to associated with this row
    # @param indent      if true, indent the first column
    # @return None
    def _write_row(self, msg, count, total, css_class=None, indent=False):
        pct = self._pct(count, total).strip()

        self._gen_tr_open(css_class)
        self._gen_td(msg, css_class='ver-summary-col1', indent=indent)
        self._gen_td(count, css_class='ver-summary-col2')
        self._gen_td(pct, css_class='ver-summary-col3')
        self._gen_tr_close()



# TODO rename for compatibility with MVC
# TODO move cfg, status, storage to model.cfg, etc

## holds configuration information
cfg = None

## holds information that spans multiple runs
status = None

## holds stored information about all test runs
storage = None

# TODO move proto, trace, summary to ctrl.proto, etc.

## holds protocol information about the current test run
proto = None

## holds trace matrix information about the current test run
trace = None

## holds summary information about the current test run
summary = None

## holds logger
logger = None

## holds test harness object
harness = None

## holds IUV object for use during self-testing
iuv = None

## holds the main window used for manual steps and verification
view = None


# -------------------
def abort(msg=None):
    # may be called by IUV, by UT, by User, or by normal operations
    if msg is None:
        line = 'abort called'
    else:
        line = f'abort called: {msg}'

    # logger may or may not be defined
    if logger is None:
        # print okay
        print(line)  # pragma: no cover
        # coverage: logger is always defined in IUV
    else:
        logger.err(line)

    import pytest

    if cfg.iuvmode:  # pragma: no cover
        # coverage: iuvmode is only set during IUV and UT runs
        # harness may or may not be defined
        if harness is None:
            # just skip the current testcase
            pytest.skip(line)
        elif harness.iuv is None:
            # UT: skip
            logger.err(f'ut: {msg}')
        else:
            harness.iuv.abort(line)
    else:
        # it's normal operation, end the pytest session
        # coverage: iuvmode is set for IUV and UT runs; never gets here
        pytest.exit(line)  # pragma: no cover



import os
import sys


# -------------------
## runs OS specific commands
# there are four recognized OS:
#  * Ubuntu
#  * Mac
#  * Windows
#  * RPi
class OsSpecific:
    ## holds the OS name
    os_name = 'unknown'

    ## holds a reference to the implementation class for the current OS
    impl = None

    # -------------------
    ## implements the Windows specific commands
    class Win:
        # -------------------
        ## constructor
        def __init__(self):
            pass

        # -------------------
        ## simple OS name
        #
        # @return string indicating OS
        def os_name(self):
            return 'win'

    # -------------------
    ## implements the Mac specific commands
    class Mac:
        # -------------------
        ## constructor
        def __init__(self):
            pass

        # -------------------
        ## simple OS name
        #
        # @return string indicating OS
        def os_name(self):
            return 'macos'

    # -------------------
    ## implements the Ubuntu specific commands
    class Ubuntu:
        # -------------------
        ## constructor
        def __init__(self):
            pass

        # -------------------
        ## simple OS name
        #
        # @return string indicating OS
        def os_name(self):
            return 'ubuntu'

    # -------------------
    ## implements the Raspberry Pi specific commands
    class RaspberryPi:
        # -------------------
        ## constructor
        def __init__(self):
            pass

        # -------------------
        ## report current OS
        #
        # @return string indicating OS
        def os_name(self):
            return 'rpi'

    # -------------------
    ## initialize
    #
    # selects the current platform and sets impl to it
    # @return None
    @classmethod
    def init(cls):
        ## holds the implementation class
        if os.path.isfile('/sys/firmware/devicetree/base/model'):
            cls.impl = OsSpecific.RaspberryPi()
        elif sys.platform == 'win32':
            cls.impl = OsSpecific.Win()
        elif sys.platform == 'darwin':
            cls.impl = OsSpecific.Mac()
        elif sys.platform == 'linux':
            cls.impl = OsSpecific.Ubuntu()
        else:
            print(f'ERR: unrecognized OS: "{sys.platform}"')
            sys.stdout.flush()
            sys.exit(1)

        ## holds the simple OS name: win, ubuntu, macOS, rpi
        cls.os_name = cls.impl.os_name()
